# -*- coding: utf-8 -*-
"""Third party libraries and modules bundled with PyVISA.

This file is part of PyVISA.

:copyright: 2014-2022 by PyVISA Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""



# -*- coding: utf-8 -*-
"""High level wrapper for USB resources.

This file is part of PyVISA.

:copyright: 2014-2022 by PyVISA Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""
from .. import attributes, constants
from ..attributes import Attribute
from .messagebased import ControlRenMixin, MessageBasedResource


class USBCommon(MessageBasedResource):
    """Common class for USB resources."""

    #: USB interface number used by the given session.
    interface_number: Attribute[int] = attributes.AttrVI_ATTR_USB_INTFC_NUM()

    #: USB serial number of this device.
    serial_number: Attribute[str] = attributes.AttrVI_ATTR_USB_SERIAL_NUM()

    #: USB protocol used by this USB interface.
    usb_protocol: Attribute[int] = attributes.AttrVI_ATTR_USB_PROTOCOL()

    #: Maximum size of data that will be stored by any given USB interrupt.
    maximum_interrupt_size: Attribute[int] = attributes.AttrVI_ATTR_USB_MAX_INTR_SIZE()

    #: Manufacturer name.
    manufacturer_name: Attribute[str] = attributes.AttrVI_ATTR_MANF_NAME()

    #: Manufacturer identification number of the device.
    manufacturer_id: Attribute[int] = attributes.AttrVI_ATTR_MANF_ID()

    #: Model name of the device.
    model_name: Attribute[str] = attributes.AttrVI_ATTR_MODEL_NAME()

    #: Model code for the device.
    model_code: Attribute[int] = attributes.AttrVI_ATTR_MODEL_CODE()


@MessageBasedResource.register(constants.InterfaceType.usb, "INSTR")
class USBInstrument(ControlRenMixin, USBCommon):
    """USB INSTR resources USB::manufacturer ID::model code::serial number

    More complex resource names can be specified with the following grammar:
        USB[board]::manufacturer ID::model code::serial number[::USB interface number][::INSTR]

    Do not instantiate directly, use
    :meth:`pyvisa.highlevel.ResourceManager.open_resource`.

    """

    #: Whether the device is 488.2 compliant.
    is_4882_compliant: Attribute[bool] = attributes.AttrVI_ATTR_4882_COMPLIANT()

    def control_in(
        self,
        request_type_bitmap_field: int,
        request_id: int,
        request_value: int,
        index: int,
        length: int = 0,
    ) -> bytes:
        """Performs a USB control pipe transfer from the device.

        Parameters
        ----------
        request_type_bitmap_field : int
            bmRequestType parameter of the setup stage of a USB control transfer.
        request_id : int
            bRequest parameter of the setup stage of a USB control transfer.
        request_value : int
            wValue parameter of the setup stage of a USB control transfer.
        index : int
            wIndex parameter of the setup stage of a USB control transfer.
            This is usually the index of the interface or endpoint.
        length : int
            wLength parameter of the setup stage of a USB control transfer.
            This value also specifies the size of the data buffer to receive
            the data from the optional data stage of the control transfer.

        Returns
        -------
        bytes
            The data buffer that receives the data from the optional data stage
            of the control transfer.

        """
        return self.visalib.usb_control_in(
            self.session,
            request_type_bitmap_field,
            request_id,
            request_value,
            index,
            length,
        )[0]

    def control_out(
        self,
        request_type_bitmap_field: int,
        request_id: int,
        request_value: int,
        index: int,
        data: bytes = b"",
    ):
        """Performs a USB control pipe transfer to the device.

        Parameters
        ----------
        request_type_bitmap_field : int
            bmRequestType parameter of the setup stage of a USB control transfer.
        request_id : int
            bRequest parameter of the setup stage of a USB control transfer.
        request_value : int
            wValue parameter of the setup stage of a USB control transfer.
        index : int
            wIndex parameter of the setup stage of a USB control transfer.
            This is usually the index of the interface or endpoint.
        data : bytes
            The data buffer that sends the data in the optional data stage of
            the control transfer.

        """
        return self.visalib.usb_control_out(
            self.session,
            request_type_bitmap_field,
            request_id,
            request_value,
            index,
            data,
        )


@MessageBasedResource.register(constants.InterfaceType.usb, "RAW")
class USBRaw(USBCommon):
    """USB RAW resources: USB::manufacturer ID::model code::serial number::RAW

    More complex resource names can be specified with the following grammar:
        USB[board]::manufacturer ID::model code::serial number[::USB interface number]::RAW

    Do not instantiate directly, use
    :meth:`pyvisa.highlevel.ResourceManager.open_resource`.

    """



# -*- coding: utf-8 -*-
"""Defines VPP 4.3.2 wrapping functions using ctypes, adding signatures to the library.

This file is part of PyVISA.

:copyright: 2014-2022 by PyVISA Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""
import warnings
from contextlib import contextmanager
from ctypes import (
    POINTER,
    byref,
    c_double,
    c_long,
    c_void_p,
    c_wchar_p,
    create_string_buffer,
)
from functools import update_wrapper
from threading import Lock
from typing import Any, Callable, Optional, Tuple

from pyvisa import attributes, constants, ctwrapper, typing
from pyvisa.highlevel import ResourceInfo

from . import types
from .types import (
    FUNCTYPE,
    ViAccessMode,
    ViAChar,
    ViAddr,
    ViAttr,
    ViAttrState,
    ViAUInt8,
    ViAUInt16,
    ViAUInt32,
    ViAUInt64,
    ViBoolean,
    ViBuf,
    ViBusAddress,
    ViBusAddress64,
    ViBusSize,
    ViEvent,
    ViEventFilter,
    ViEventType,
    ViFindList,
    ViHndlr,
    ViInt16,
    ViJobId,
    ViKeyId,
    ViObject,
    ViPAddr,
    ViPBuf,
    ViPBusAddress,
    ViPEvent,
    ViPEventType,
    ViPFindList,
    ViPJobId,
    ViPSession,
    ViPUInt8,
    ViPUInt16,
    ViPUInt32,
    ViPUInt64,
    ViRsrc,
    ViSession,
    ViStatus,
    ViString,
    ViUInt8,
    ViUInt16,
    ViUInt32,
    ViUInt64,
    buffer_to_text,
)

visa_functions = [
    "assert_interrupt_signal",
    "assert_trigger",
    "assert_utility_signal",
    "buffer_read",
    "buffer_write",
    "clear",
    "close",
    "disable_event",
    "discard_events",
    "enable_event",
    "_find_next",
    "_find_resources",
    "flush",
    "get_attribute",
    "gpib_command",
    "gpib_control_atn",
    "gpib_control_ren",
    "gpib_pass_control",
    "gpib_send_ifc",
    "in_16",
    "in_32",
    "in_8",
    "install_handler",
    "lock",
    "map_address",
    "map_trigger",
    "memory_allocation",
    "memory_free",
    "move",
    "move_asynchronously",
    "move_in_16",
    "move_in_32",
    "move_in_8",
    "move_out_16",
    "move_out_32",
    "move_out_8",
    "open",
    "open_default_resource_manager",
    "out_16",
    "out_32",
    "out_8",
    "parse_resource",
    "parse_resource_extended",
    "peek_16",
    "peek_32",
    "peek_8",
    "poke_16",
    "poke_32",
    "poke_8",
    "read",
    "read_to_file",
    "read_stb",
    "set_attribute",
    "set_buffer",
    "status_description",
    "terminate",
    "uninstall_handler",
    "unlock",
    "unmap_address",
    "unmap_trigger",
    "usb_control_in",
    "usb_control_out",
    "vxi_command_query",
    "wait_on_event",
    "write",
    "write_asynchronously",
    "write_from_file",
    "in_64",
    "move_in_64",
    "out_64",
    "move_out_64",
    "poke_64",
    "peek_64",
]

__all__ = ["visa_functions", "set_signatures"] + visa_functions

VI_SPEC_VERSION = 0x00300000

#: Global lock to ensure that we cannot have one thread change the type while
#: another is trying to interact with VISA
ViHndlr_lock = Lock()


@contextmanager
def set_user_handle_type(library, user_handle: Any):
    """Set the type of the user handle to install and uninstall handler signature.

    Parameters
    ----------
    library :
        The visa library wrapped by ctypes.
    user_handle :
        User handle used when registering an event handler. Use None for a void_p.

    """
    with ViHndlr_lock:
        # Actually, it's not necessary to change ViHndlr *globally*.  However,
        # I don't want to break symmetry too much with all the other VPP43
        # routines.
        global ViHndlr

        if user_handle is None:
            user_handle_p = c_void_p
        else:
            user_handle_p = POINTER(type(user_handle))  # type: ignore

        ViHndlr = FUNCTYPE(ViStatus, ViSession, ViEventType, ViEvent, user_handle_p)
        library.viInstallHandler.argtypes = [
            ViSession,
            ViEventType,
            ViHndlr,
            user_handle_p,
        ]
        library.viUninstallHandler.argtypes = [
            ViSession,
            ViEventType,
            ViHndlr,
            user_handle_p,
        ]
        yield


def set_signatures(
    library, errcheck: Optional[Callable[[int, Callable, tuple], int]] = None
):
    """Set the signatures of most visa functions in the library.

    All instrumentation related functions are specified here.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        The visa library wrapped by ctypes.
    errcheck : Optional[Callable[[int, Callable, tuple], int]]
        Error checking callable used for visa functions that return ViStatus.
        It should be take three arguments (result, func, arguments).
        See errcheck in ctypes.

    """
    # Somehow hasattr(library, '_functions') segfaults in cygwin (See #131)
    if "_functions" not in dir(library):
        library._functions = []
        library._functions_failed = []

    def _applier(restype, errcheck_):
        def _internal(function_name, argtypes, required=False):
            try:
                set_signature(library, function_name, argtypes, restype, errcheck_)
                library._functions.append(function_name)
            except AttributeError:
                library._functions_failed.append(function_name)
                if required:
                    raise

        return _internal

    # Visa functions with ViStatus return code
    apply = _applier(ViStatus, errcheck)
    apply("viAssertIntrSignal", [ViSession, ViInt16, ViUInt32])
    apply("viAssertTrigger", [ViSession, ViUInt16])
    apply("viAssertUtilSignal", [ViSession, ViUInt16])
    apply("viBufRead", [ViSession, ViPBuf, ViUInt32, ViPUInt32])
    apply("viBufWrite", [ViSession, ViBuf, ViUInt32, ViPUInt32])
    apply("viClear", [ViSession])
    apply("viClose", [ViObject])
    apply("viDisableEvent", [ViSession, ViEventType, ViUInt16])
    apply("viDiscardEvents", [ViSession, ViEventType, ViUInt16])
    apply("viEnableEvent", [ViSession, ViEventType, ViUInt16, ViEventFilter])
    apply("viFindNext", [ViSession, ViAChar])
    apply("viFindRsrc", [ViSession, ViString, ViPFindList, ViPUInt32, ViAChar])
    apply("viFlush", [ViSession, ViUInt16])
    apply("viGetAttribute", [ViObject, ViAttr, c_void_p])
    apply("viGpibCommand", [ViSession, ViBuf, ViUInt32, ViPUInt32])
    apply("viGpibControlATN", [ViSession, ViUInt16])
    apply("viGpibControlREN", [ViSession, ViUInt16])
    apply("viGpibPassControl", [ViSession, ViUInt16, ViUInt16])
    apply("viGpibSendIFC", [ViSession])

    apply("viIn8", [ViSession, ViUInt16, ViBusAddress, ViPUInt8])
    apply("viIn16", [ViSession, ViUInt16, ViBusAddress, ViPUInt16])
    apply("viIn32", [ViSession, ViUInt16, ViBusAddress, ViPUInt32])
    apply("viIn64", [ViSession, ViUInt16, ViBusAddress, ViPUInt64])

    apply("viIn8Ex", [ViSession, ViUInt16, ViBusAddress64, ViPUInt8])
    apply("viIn16Ex", [ViSession, ViUInt16, ViBusAddress64, ViPUInt16])
    apply("viIn32Ex", [ViSession, ViUInt16, ViBusAddress64, ViPUInt32])
    apply("viIn64Ex", [ViSession, ViUInt16, ViBusAddress64, ViPUInt64])

    apply("viInstallHandler", [ViSession, ViEventType, ViHndlr, ViAddr])
    apply("viLock", [ViSession, ViAccessMode, ViUInt32, ViKeyId, ViAChar])
    apply(
        "viMapAddress",
        [ViSession, ViUInt16, ViBusAddress, ViBusSize, ViBoolean, ViAddr, ViPAddr],
    )
    apply("viMapTrigger", [ViSession, ViInt16, ViInt16, ViUInt16])
    apply("viMemAlloc", [ViSession, ViBusSize, ViPBusAddress])
    apply("viMemFree", [ViSession, ViBusAddress])
    apply(
        "viMove",
        [
            ViSession,
            ViUInt16,
            ViBusAddress,
            ViUInt16,
            ViUInt16,
            ViBusAddress,
            ViUInt16,
            ViBusSize,
        ],
    )
    apply(
        "viMoveAsync",
        [
            ViSession,
            ViUInt16,
            ViBusAddress,
            ViUInt16,
            ViUInt16,
            ViBusAddress,
            ViUInt16,
            ViBusSize,
            ViPJobId,
        ],
    )

    apply("viMoveIn8", [ViSession, ViUInt16, ViBusAddress, ViBusSize, ViAUInt8])
    apply("viMoveIn16", [ViSession, ViUInt16, ViBusAddress, ViBusSize, ViAUInt16])
    apply("viMoveIn32", [ViSession, ViUInt16, ViBusAddress, ViBusSize, ViAUInt32])
    apply("viMoveIn64", [ViSession, ViUInt16, ViBusAddress, ViBusSize, ViAUInt64])

    apply("viMoveIn8Ex", [ViSession, ViUInt16, ViBusAddress64, ViBusSize, ViAUInt8])
    apply("viMoveIn16Ex", [ViSession, ViUInt16, ViBusAddress64, ViBusSize, ViAUInt16])
    apply("viMoveIn32Ex", [ViSession, ViUInt16, ViBusAddress64, ViBusSize, ViAUInt32])
    apply("viMoveIn64Ex", [ViSession, ViUInt16, ViBusAddress64, ViBusSize, ViAUInt64])

    apply("viMoveOut8", [ViSession, ViUInt16, ViBusAddress, ViBusSize, ViAUInt8])
    apply("viMoveOut16", [ViSession, ViUInt16, ViBusAddress, ViBusSize, ViAUInt16])
    apply("viMoveOut32", [ViSession, ViUInt16, ViBusAddress, ViBusSize, ViAUInt32])
    apply("viMoveOut64", [ViSession, ViUInt16, ViBusAddress, ViBusSize, ViAUInt64])

    apply("viMoveOut8Ex", [ViSession, ViUInt16, ViBusAddress64, ViBusSize, ViAUInt8])
    apply("viMoveOut16Ex", [ViSession, ViUInt16, ViBusAddress64, ViBusSize, ViAUInt16])
    apply("viMoveOut32Ex", [ViSession, ViUInt16, ViBusAddress64, ViBusSize, ViAUInt32])
    apply("viMoveOut64Ex", [ViSession, ViUInt16, ViBusAddress64, ViBusSize, ViAUInt64])

    apply(
        "viOpen", [ViSession, ViRsrc, ViAccessMode, ViUInt32, ViPSession], required=True
    )

    apply("viOpenDefaultRM", [ViPSession], required=True)

    apply("viOut8", [ViSession, ViUInt16, ViBusAddress, ViUInt8])
    apply("viOut16", [ViSession, ViUInt16, ViBusAddress, ViUInt16])
    apply("viOut32", [ViSession, ViUInt16, ViBusAddress, ViUInt32])
    apply("viOut64", [ViSession, ViUInt16, ViBusAddress, ViUInt64])

    apply("viOut8Ex", [ViSession, ViUInt16, ViBusAddress64, ViUInt8])
    apply("viOut16Ex", [ViSession, ViUInt16, ViBusAddress64, ViUInt16])
    apply("viOut32Ex", [ViSession, ViUInt16, ViBusAddress64, ViUInt32])
    apply("viOut64Ex", [ViSession, ViUInt16, ViBusAddress64, ViUInt64])

    apply("viParseRsrc", [ViSession, ViRsrc, ViPUInt16, ViPUInt16])
    apply(
        "viParseRsrcEx",
        [ViSession, ViRsrc, ViPUInt16, ViPUInt16, ViAChar, ViAChar, ViAChar],
    )

    apply("viRead", [ViSession, ViPBuf, ViUInt32, ViPUInt32])
    apply("viReadAsync", [ViSession, ViPBuf, ViUInt32, ViPJobId])
    apply("viReadSTB", [ViSession, ViPUInt16])
    apply("viReadToFile", [ViSession, ViString, ViUInt32, ViPUInt32])

    apply("viSetAttribute", [ViObject, ViAttr, ViAttrState])
    apply("viSetBuf", [ViSession, ViUInt16, ViUInt32])

    apply("viStatusDesc", [ViObject, ViStatus, ViAChar])
    apply("viTerminate", [ViSession, ViUInt16, ViJobId])
    apply("viUninstallHandler", [ViSession, ViEventType, ViHndlr, ViAddr])
    apply("viUnlock", [ViSession])
    apply("viUnmapAddress", [ViSession])
    apply("viUnmapTrigger", [ViSession, ViInt16, ViInt16])
    apply(
        "viUsbControlIn",
        [ViSession, ViInt16, ViInt16, ViUInt16, ViUInt16, ViUInt16, ViPBuf, ViPUInt16],
    )
    apply(
        "viUsbControlOut",
        [ViSession, ViInt16, ViInt16, ViUInt16, ViUInt16, ViUInt16, ViPBuf],
    )

    # The following "V" routines are *not* implemented in PyVISA, and will
    # never be: viVPrintf, viVQueryf, viVScanf, viVSPrintf, viVSScanf

    apply("viVxiCommandQuery", [ViSession, ViUInt16, ViUInt32, ViPUInt32])
    apply("viWaitOnEvent", [ViSession, ViEventType, ViUInt32, ViPEventType, ViPEvent])
    apply("viWrite", [ViSession, ViBuf, ViUInt32, ViPUInt32])
    apply("viWriteAsync", [ViSession, ViBuf, ViUInt32, ViPJobId])
    apply("viWriteFromFile", [ViSession, ViString, ViUInt32, ViPUInt32])

    # Functions that return void.
    apply = _applier(None, None)
    apply("viPeek8", [ViSession, ViAddr, ViPUInt8])
    apply("viPeek16", [ViSession, ViAddr, ViPUInt16])
    apply("viPeek32", [ViSession, ViAddr, ViPUInt32])
    apply("viPeek64", [ViSession, ViAddr, ViPUInt64])

    apply("viPoke8", [ViSession, ViAddr, ViUInt8])
    apply("viPoke16", [ViSession, ViAddr, ViUInt16])
    apply("viPoke32", [ViSession, ViAddr, ViUInt32])
    apply("viPoke64", [ViSession, ViAddr, ViUInt64])


def set_signature(
    library,
    function_name: str,
    argtypes: tuple,
    restype,
    errcheck: Optional[Callable[[int, Callable, tuple], int]],
):
    """Set the signature of single function in a library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    function_name : str
        Name of the function as appears in the header file.
    argtypes : tuple
        ctypes types to specify the argument types that the function accepts.
    restype :
        A ctypes type to specify the result type of the foreign function.
        Use None for void, a function not returning anything.
    errcheck : Optional[Callable[[int, Callable, tuple], int]]
        Error checking callable used for visa functions that return ViStatus.
        It should be take three arguments (result, func, arguments).
        See errcheck in ctypes.

    """

    func = getattr(library, function_name)
    func.argtypes = argtypes
    if restype is not None:
        func.restype = restype
    if errcheck is not None:
        func.errcheck = errcheck


# The VPP-4.3.2 routines

# Usually, there is more than one way to pass parameters to ctypes calls.  The
# ctypes policy used in this code goes as follows:
#
# * Null pointers are passed as "None" rather than "0".  This is a little bit
#   unfortunate, since the VPP specification calls this "VI_NULL", but I can't
#   use "VI_NULL" since it's an integer and may not be compatible with a
#   pointer type (don't know whether this is really dangerous).
#
# * Strings must have been created with "create_string_buffer" and are passed
#   without any further conversion; they stand in the parameter list as is.
#   The same applies to pseudo-string types as ViRsrc or VuBuf.  Their Pythonic
#   counterpats are strings as well.
#
# * All other types are explicitly cast using the types defined by ctypes'
#   "restype".
#
# Further notes:
#
# * The following Python routines take and give handles as ctypes objects.
#   Since the user shouldn't be interested in handle values anyway, I see no
#   point in converting them to Python strings or integers.
#
# * All other parameters are natural Python types, i.e. strings (may contain
#   binary data) and integers.  The same is true for return values.
#
# * The original VPP function signatures cannot be realised in Python, at least
#   not in a sensible way, because a) Python has no real call-by-reference, and
#   b) Python allows for more elegant solutions, e.g. using len(buffer) instead
#   of a separate "count" parameter, or using tuples as return values.
#
#   Therefore, all function signatures have been carefully adjusted.  I think
#   this is okay, since the original standard must be adopted to at least C and
#   Visual Basic anyway, with slight modifications.  I also made the function
#   names and parameters more legible, but in a way that it's perfectly clear
#   which original function is meant.
#
#   The important thing is that the semantics of functions and parameters are
#   totally intact, and the inner order of parameters, too.  There is a 1:1
#   mapping.


def assert_interrupt_signal(library, session, mode, status_id):
    """Asserts the specified interrupt or signal.

    Corresponds to viAssertIntrSignal function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    mode : constants.AssertSignalInterrupt
        How to assert the interrupt.
    status_id : int
        Status value to be presented during an interrupt acknowledge cycle.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viAssertIntrSignal(session, mode, status_id)


def assert_trigger(library, session, protocol):
    """Assert software or hardware trigger.

    Corresponds to viAssertTrigger function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    protocol : constants.TriggerProtocol
        Trigger protocol to use during assertion.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viAssertTrigger(session, protocol)


def assert_utility_signal(library, session, line):
    """Assert or deassert the specified utility bus signal.

    Corresponds to viAssertUtilSignal function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    line : constants.UtilityBusSignal
        Specifies the utility bus signal to assert.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viAssertUtilSignal(session, line)


def buffer_read(library, session, count):
    """Reads data through the use of a formatted I/O read buffer.

    The data can be read from a device or an interface.

    Corresponds to viBufRead function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    count : int
        Number of bytes to be read.

    Returns
    -------
    dbytes
        Data read
    constants.StatusCode
        Return value of the library call.

    """
    buffer = create_string_buffer(count)
    return_count = ViUInt32()
    ret = library.viBufRead(session, buffer, count, byref(return_count))
    return buffer.raw[: return_count.value], ret


def buffer_write(library, session, data):
    """Writes data to a formatted I/O write buffer synchronously.

    Corresponds to viBufWrite function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    data : bytes
        Data to be written.

    Returns
    -------
    int
        number of written bytes
    constants.StatusCode
        return value of the library call.

    """
    return_count = ViUInt32()
    # [ViSession, ViBuf, ViUInt32, ViPUInt32]
    ret = library.viBufWrite(session, data, len(data), byref(return_count))
    return return_count.value, ret


def clear(library, session):
    """Clears a device.

    Corresponds to viClear function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viClear(session)


def close(library, session):
    """Closes the specified session, event, or find list.

    Corresponds to viClose function of the VISA library.

    Parameters
    ---------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : Union[VISASession, VISAEventContext, VISARMSession]
        Unique logical identifier to a session, event, resource manager.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viClose(session)


def disable_event(library, session, event_type, mechanism):
    """Disable notification for an event type(s) via the specified mechanism(s).

    Corresponds to viDisableEvent function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    event_type : constants.EventType
        Event type.
    mechanism : constants.EventMechanism
        Event handling mechanisms to be disabled.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viDisableEvent(session, event_type, mechanism)


def discard_events(library, session, event_type, mechanism):
    """Discard event occurrences for a given type and mechanisms in a session.

    Corresponds to viDiscardEvents function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    event_type : constans.EventType
        Logical event identifier.
    mechanism : constants.EventMechanism
        Specifies event handling mechanisms to be discarded.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viDiscardEvents(session, event_type, mechanism)


def enable_event(library, session, event_type, mechanism, context=None):
    """Enable event occurrences for specified event types and mechanisms in a session.

    Corresponds to viEnableEvent function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    event_type : constants.EventType
        Logical event identifier.
    mechanism : constants.EventMechanism
        Specifies event handling mechanisms to be enabled.
    context : None, optional
        Unused parameter...

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    if context is None:
        context = constants.VI_NULL
    elif context != constants.VI_NULL:
        warnings.warn("In enable_event, context will be set VI_NULL.")
        context = constants.VI_NULL  # according to spec VPP-4.3, section 3.7.3.1
    return library.viEnableEvent(session, event_type, mechanism, context)


def _find_next(library, find_list: ViFindList) -> Tuple[str, constants.StatusCode]:
    """Get next resource from the list of resources.

    The list of resources should be obtained from a previous call to find_resources().

    Corresponds to viFindNext function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    find_list :
        Describes a find list. This parameter must be created by find_resources().

    Returns
    -------
    str
        String identifying the location of a device
    constants.StatusCode
        Return value of the library call.

    """
    instrument_description = create_string_buffer(constants.VI_FIND_BUFLEN)
    ret = library.viFindNext(find_list, instrument_description)
    return buffer_to_text(instrument_description), ret


def _find_resources(library, session: typing.VISARMSession, query: str):
    """Queries VISA to locate the resources associated with a specified interface.

    Corresponds to viFindRsrc function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : typing.VISARMSession
        Unique logical identifier to the ResourceManger session
        (unused, just to uniform signatures).
    query : str
        A regular expression followed by an optional logical expression.
        Use '?*' for all.

    Returns
    -------
    ViFindList
        Opaque object to pass to `_find_next` to access the other devices
        resource name.
    int
        Number of identified devices.
    str
        Resource name of the first identified device
    constants.StatusCode
        Return value of the library call.

    """
    find_list = ViFindList()
    return_counter = ViUInt32()
    instrument_description = create_string_buffer(constants.VI_FIND_BUFLEN)

    ret = library.viFindRsrc(
        session, query, byref(find_list), byref(return_counter), instrument_description
    )
    return find_list, return_counter.value, buffer_to_text(instrument_description), ret


def flush(library, session, mask):
    """Retrieves the state of an attribute.

    Corresponds to viGetAttribute function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : Union[VISASession, VISAEventContext]
        Unique logical identifier to a session, event, or find list.
    attribute : Union[constants.ResourceAttribute, constants.EventAttribute]
        Resource or event attribute for which the state query is made.

    Returns
    -------
    Any
        State of the queried attribute for a specified resource
    constants.StatusCode
        Return value of the library call.

    """
    return library.viFlush(session, mask)


def get_attribute(library, session, attribute):
    """Retrieves the state of an attribute.

    Corresponds to viGetAttribute function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : Union[VISASession, VISAEventContext]
        Unique logical identifier to a session, event, or find list.
    attribute : Union[constants.ResourceAttribute, constants.EventAttribute]
        Resource or event attribute for which the state query is made.

    Returns
    -------
    Any
        State of the queried attribute for a specified resource
    constants.StatusCode
        Return value of the library call.

    """
    attr = attributes.AttributesByID[attribute]
    datatype = getattr(types, attr.visa_type)
    if datatype == ViString:
        attribute_state = create_string_buffer(256)
        ret = library.viGetAttribute(session, attribute, attribute_state)
        return buffer_to_text(attribute_state), ret
    # There are only 2 buffer attribute, the one we do not handle if the one
    # to async read that is handled at a higher since we pass the buffer ourself
    elif datatype == ViBuf:
        if attr.visa_name == "VI_ATTR_USB_RECV_INTR_DATA":
            length = get_attribute(
                library, session, constants.VI_ATTR_USB_RECV_INTR_SIZE
            )
            attribute_state = (ViUInt8 * length)()
            ret = library.viGetAttribute(session, attribute, byref(attribute_state))
            return list(attribute_state), ret
        else:
            raise AttributeError("%s cannot be accessed directly" % attr.visa_name)
    else:
        attribute_state = datatype()
        ret = library.viGetAttribute(session, attribute, byref(attribute_state))
        return attribute_state.value, ret


def gpib_command(library, session, data):
    """Write GPIB command bytes on the bus.

    Corresponds to viGpibCommand function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    data : bytes
        Data to write.

    Returns
    -------
    int
        Number of written bytes
    constants.StatusCode
        Return value of the library call.

    """
    return_count = ViUInt32()

    # [ViSession, ViBuf, ViUInt32, ViPUInt32]
    ret = library.viGpibCommand(session, data, len(data), byref(return_count))
    return return_count.value, ret


def gpib_control_atn(library, session, mode):
    """Specifies the state of the ATN line and the local active controller state.

    Corresponds to viGpibControlATN function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    mode : constants.ATNLineOperation
        State of the ATN line and optionally the local active controller state.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viGpibControlATN(session, mode)


def gpib_control_ren(library, session, mode):
    """Controls the state of the GPIB Remote Enable (REN) interface line.

    Optionally the remote/local state of the device can also be set.

    Corresponds to viGpibControlREN function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    mode : constants.RENLineOperation
        State of the REN line and optionally the device remote/local state.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viGpibControlREN(session, mode)


def gpib_pass_control(library, session, primary_address, secondary_address):
    """Tell a GPIB device to become controller in charge (CIC).

    Corresponds to viGpibPassControl function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    primary_address : int
        Primary address of the GPIB device to which you want to pass control.
    secondary_address : int
        Secondary address of the targeted GPIB device.
        If the targeted device does not have a secondary address, this parameter
        should contain the value Constants.VI_NO_SEC_ADDR.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viGpibPassControl(session, primary_address, secondary_address)


def gpib_send_ifc(library, session):
    """Pulse the interface clear line (IFC) for at least 100 microseconds.

    Corresponds to viGpibSendIFC function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viGpibSendIFC(session)


def in_8(library, session, space, offset, extended=False):
    """Reads in an 8-bit value from the specified memory space and offset.

    Corresponds to viIn8* function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Specifies the address space.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    extended : bool, optional
        Use 64 bits offset independent of the platform, False by default.

    Returns
    -------
    int
        Data read from memory
    constants.StatusCode
        Return value of the library call.

    """
    value_8 = ViUInt8()
    if extended:
        ret = library.viIn8Ex(session, space, offset, byref(value_8))
    else:
        ret = library.viIn8(session, space, offset, byref(value_8))
    return value_8.value, ret


def in_16(library, session, space, offset, extended=False):
    """Reads in an 16-bit value from the specified memory space and offset.

    Corresponds to viIn16* function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Specifies the address space.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    extended : bool, optional
        Use 64 bits offset independent of the platform, False by default.

    Returns
    -------
    int
        Data read from memory
    constants.StatusCode
        Return value of the library call.

    """
    value_16 = ViUInt16()
    if extended:
        ret = library.viIn16Ex(session, space, offset, byref(value_16))
    else:
        ret = library.viIn16(session, space, offset, byref(value_16))
    return value_16.value, ret


def in_32(library, session, space, offset, extended=False):
    """Reads in an 32-bit value from the specified memory space and offset.

    Corresponds to viIn32* function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Specifies the address space.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    extended : bool, optional
        Use 64 bits offset independent of the platform, False by default.

    Returns
    -------
    int
        Data read from memory
    constants.StatusCode
        Return value of the library call.

    """
    value_32 = ViUInt32()
    if extended:
        ret = library.viIn32Ex(session, space, offset, byref(value_32))
    else:
        ret = library.viIn32(session, space, offset, byref(value_32))
    return value_32.value, ret


def in_64(library, session, space, offset, extended=False):
    """Reads in an 64-bit value from the specified memory space and offset.

    Corresponds to viIn64* function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Specifies the address space.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    extended : bool, optional
        Use 64 bits offset independent of the platform, False by default.

    Returns
    -------
    int
        Data read from memory
    constants.StatusCode
        Return value of the library call.

    """
    value_64 = ViUInt64()
    if extended:
        ret = library.viIn64Ex(session, space, offset, byref(value_64))
    else:
        ret = library.viIn64(session, space, offset, byref(value_64))
    return value_64.value, ret


def install_handler(
    library, session, event_type, handler, user_handle: Any
) -> Tuple[typing.VISAHandler, Any, Any, constants.StatusCode]:
    """Install handlers for event callbacks.

    Corresponds to viInstallHandler function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    event_type : constants.EventType
        Logical event identifier.
    handler : VISAHandler
        Reference to a handler to be installed by a client application.
    user_handle : Any
        Value specified by an application that can be used for identifying
        handlers uniquely for an event type.

    Returns
    -------
    handler : VISAHandler
        Handler to be installed by a client application.
    converted_user_handle :
        Converted user handle to match the underlying library. This version
        of the handle should be used in further call to the library.
    converted_handler :
        Converted version of the handler satisfying to backend library.
    status_code : constants.StatusCode
        Return value of the library call

    """
    # Should be Optional[_CData] but that type cannot be imported
    converted_user_handle: object = None
    if user_handle is not None:
        if isinstance(user_handle, int):
            converted_user_handle = c_long(user_handle)
        elif isinstance(user_handle, float):
            converted_user_handle = c_double(user_handle)
        elif isinstance(user_handle, str):
            converted_user_handle = c_wchar_p(user_handle)
        elif isinstance(user_handle, list):
            for element in user_handle:
                if not isinstance(element, int):
                    # Mypy cannot track the fact that the list has to contain float
                    converted_user_handle = (c_double * len(user_handle))(  # type: ignore
                        *tuple(user_handle)
                    )
                    break
            else:
                converted_user_handle = (c_long * len(user_handle))(*tuple(user_handle))
        else:
            try:
                # check if it is already a ctypes
                byref(user_handle)
                converted_user_handle = user_handle
            except TypeError:
                raise TypeError(
                    "Type not allowed as user handle: %s" % type(user_handle)
                )

    with set_user_handle_type(library, converted_user_handle):
        if ctwrapper.WRAP_HANDLER:
            # Wrap the handler to provide a non-wrapper specific interface
            def handler_wrapper(
                ctype_session, ctype_event_type, ctype_event_context, ctype_user_handle
            ):
                handler(
                    ctype_session.value,
                    ctype_event_type,
                    ctype_event_context.value,
                    ctype_user_handle.contents
                    if ctype_user_handle
                    else ctype_user_handle,
                )
                return 0

            update_wrapper(handler_wrapper, handler)
        else:
            handler_wrapper = handler

        converted_handler = ViHndlr(handler_wrapper)
        if user_handle is None:
            ret = library.viInstallHandler(session, event_type, converted_handler, None)
        else:
            ret = library.viInstallHandler(
                session,
                event_type,
                converted_handler,
                byref(converted_user_handle),  # type: ignore
            )

    return handler, converted_user_handle, converted_handler, ret


def lock(library, session, lock_type, timeout, requested_key=None):
    """Establishes an access mode to the specified resources.

    Corresponds to viLock function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    lock_type : constants.Lock
        Specifies the type of lock requested.
    timeout : int
        Absolute time period (in milliseconds) that a resource waits to get
        unlocked by the locking session before returning an error.
    requested_key : Optional[str], optional
        Requested locking key in the case of a shared lock. For an exclusive
        lock it should be None.

    Returns
    -------
    Optional[str]
        Key that can then be passed to other sessions to share the lock, or
        None for an exclusive lock.
    constants.StatusCode
        Return value of the library call.

    """
    if lock_type == constants.AccessModes.exclusive_lock:
        requested_key = None
        access_key = None
    else:
        access_key = create_string_buffer(256)
    ret = library.viLock(session, lock_type, timeout, requested_key, access_key)
    if access_key is None:
        return None, ret
    else:
        return access_key.value, ret


def map_address(
    library, session, map_space, map_base, map_size, access=False, suggested=None
):
    """Maps the specified memory space into the process's address space.

    Corresponds to viMapAddress function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    map_space : constants.AddressSpace
        Specifies the address space to map.
    map_base : int
        Offset (in bytes) of the memory to be mapped.
    map_size : int
        Amount of memory to map (in bytes).
    access : False
        Unused parameter.
    suggested : Optional[int], optional
        If not None, the operating system attempts to map the memory to the
        address specified. There is no guarantee, however, that the memory
        will be mapped to that address. This operation may map the memory
        into an address region different from the suggested one.

    Returns
    -------
    int
        Address in your process space where the memory was mapped
    constants.StatusCode
        Return value of the library call.

    """
    if access is False:
        access = constants.VI_FALSE
    elif access != constants.VI_FALSE:
        warnings.warn("In enable_event, context will be set VI_NULL.")
        access = constants.VI_FALSE
    address = ViAddr()
    ret = library.viMapAddress(
        session, map_space, map_base, map_size, access, suggested, byref(address)
    )
    return address, ret


def map_trigger(library, session, trigger_source, trigger_destination, mode):
    """Map the specified trigger source line to the specified destination line.

    Corresponds to viMapTrigger function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    trigger_source : constants.InputTriggerLine
        Source line from which to map.
    trigger_destination : constants.OutputTriggerLine
        Destination line to which to map.
    mode : None, optional
        Always None for this version of the VISA specification.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viMapTrigger(session, trigger_source, trigger_destination, mode)


def memory_allocation(library, session, size, extended=False):
    """Allocate memory from a resource's memory region.

    Corresponds to viMemAlloc* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    size : int
        Specifies the size of the allocation.
    extended : bool, optional
        Use 64 bits offset independent of the platform.

    Returns
    -------
    int
        offset of the allocated memory
    constants.StatusCode
        Return value of the library call.

    """
    offset = ViBusAddress()
    if extended:
        ret = library.viMemAllocEx(session, size, byref(offset))
    else:
        ret = library.viMemAlloc(session, size, byref(offset))
    return offset, ret


def memory_free(library, session, offset, extended=False):
    """Frees memory previously allocated using the memory_allocation() operation.

    Corresponds to viMemFree* function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    offset : int
        Offset of the memory to free.
    extended : bool, optional
        Use 64 bits offset independent of the platform.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    if extended:
        return library.viMemFreeEx(session, offset)
    else:
        return library.viMemFree(session, offset)


def move(
    library,
    session,
    source_space,
    source_offset,
    source_width,
    destination_space,
    destination_offset,
    destination_width,
    length,
):
    """Moves a block of data.

    Corresponds to viMove function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    source_space : constants.AddressSpace
        Specifies the address space of the source.
    source_offset : int
        Offset of the starting address or register from which to read.
    source_width : constants.DataWidth
        Specifies the data width of the source.
    destination_space : constants.AddressSpace
        Specifies the address space of the destination.
    destination_offset : int
        Offset of the starting address or register to which to write.
    destination_width : constants.DataWidth
        Specifies the data width of the destination.
    length: int
        Number of elements to transfer, where the data width of the
        elements to transfer is identical to the source data width.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viMove(
        session,
        source_space,
        source_offset,
        source_width,
        destination_space,
        destination_offset,
        destination_width,
        length,
    )


def move_asynchronously(
    library,
    session,
    source_space,
    source_offset,
    source_width,
    destination_space,
    destination_offset,
    destination_width,
    length,
):
    """Moves a block of data asynchronously.

    Corresponds to viMoveAsync function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    source_space : constants.AddressSpace
        Specifies the address space of the source.
    source_offset : int
        Offset of the starting address or register from which to read.
    source_width : constants.DataWidth
        Specifies the data width of the source.
    destination_space : constants.AddressSpace
        Specifies the address space of the destination.
    destination_offset : int
        Offset of the starting address or register to which to write.
    destination_width : constants.DataWidth
        Specifies the data width of the destination.
    length : int
        Number of elements to transfer, where the data width of the
        elements to transfer is identical to the source data width.

    Returns
    -------
    VISAJobID
        Job identifier of this asynchronous move operation
    constants.StatusCode
        Return value of the library call.

    """
    job_id = ViJobId()
    ret = library.viMoveAsync(
        session,
        source_space,
        source_offset,
        source_width,
        destination_space,
        destination_offset,
        destination_width,
        length,
        byref(job_id),
    )
    return job_id, ret


def move_in_8(library, session, space, offset, length, extended=False):
    """Moves an 8-bit block of data to local memory.

    Corresponds to viMoveIn8* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space from which to move the data.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    length : int
            Number of elements to transfer, where the data width of
            the elements to transfer is identical to the source data width.
    extended : bool, optional
        Use 64 bits offset independent of the platform, by default False.

    Returns
    -------
    data : List[int]
        Data read from the bus
    status_code : constants.StatusCode
        Return value of the library call.

    """
    buffer_8 = (ViUInt8 * length)()
    if extended:
        ret = library.viMoveIn8Ex(session, space, offset, length, buffer_8)
    else:
        ret = library.viMoveIn8(session, space, offset, length, buffer_8)
    return list(buffer_8), ret


def move_in_16(library, session, space, offset, length, extended=False):
    """Moves an 16-bit block of data to local memory.

    Corresponds to viMoveIn816 functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space from which to move the data.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    length : int
            Number of elements to transfer, where the data width of
            the elements to transfer is identical to the source data width.
    extended : bool, optional
        Use 64 bits offset independent of the platform, by default False.

    Returns
    -------
    data : List[int]
        Data read from the bus
    status_code : constants.StatusCode
        Return value of the library call.

    """
    buffer_16 = (ViUInt16 * length)()
    if extended:
        ret = library.viMoveIn16Ex(session, space, offset, length, buffer_16)
    else:
        ret = library.viMoveIn16(session, space, offset, length, buffer_16)

    return list(buffer_16), ret


def move_in_32(library, session, space, offset, length, extended=False):
    """Moves an 32-bit block of data to local memory.

    Corresponds to viMoveIn32* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space from which to move the data.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    length : int
            Number of elements to transfer, where the data width of
            the elements to transfer is identical to the source data width.
    extended : bool, optional
        Use 64 bits offset independent of the platform, by default False.

    Returns
    -------
    data : List[int]
        Data read from the bus
    status_code : constants.StatusCode
        Return value of the library call.

    """
    buffer_32 = (ViUInt32 * length)()
    if extended:
        ret = library.viMoveIn32Ex(session, space, offset, length, buffer_32)
    else:
        ret = library.viMoveIn32(session, space, offset, length, buffer_32)

    return list(buffer_32), ret


def move_in_64(library, session, space, offset, length, extended=False):
    """Moves an 64-bit block of data to local memory.

    Corresponds to viMoveIn8* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space from which to move the data.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    length : int
            Number of elements to transfer, where the data width of
            the elements to transfer is identical to the source data width.
    extended : bool, optional
        Use 64 bits offset independent of the platform, by default False.

    Returns
    -------
    data : List[int]
        Data read from the bus
    status_code : constants.StatusCode
        Return value of the library call.

    """
    buffer_64 = (ViUInt64 * length)()
    if extended:
        ret = library.viMoveIn64Ex(session, space, offset, length, buffer_64)
    else:
        ret = library.viMoveIn64(session, space, offset, length, buffer_64)

    return list(buffer_64), ret


def move_out_8(library, session, space, offset, length, data, extended=False):
    """Moves an 8-bit block of data from local memory.

    Corresponds to viMoveOut8* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space into which move the data.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    length : int
        Number of elements to transfer, where the data width of
        the elements to transfer is identical to the source data width.
    data : Iterable[int]
        Data to write to bus.
    extended : bool, optional
        Use 64 bits offset independent of the platform, by default False.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    converted_buffer = (ViUInt8 * length)(*tuple(data))
    if extended:
        return library.viMoveOut8Ex(session, space, offset, length, converted_buffer)
    else:
        return library.viMoveOut8(session, space, offset, length, converted_buffer)


def move_out_16(library, session, space, offset, length, data, extended=False):
    """Moves an 16-bit block of data from local memory.

    Corresponds to viMoveOut16* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space into which move the data.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    length : int
        Number of elements to transfer, where the data width of
        the elements to transfer is identical to the source data width.
    data : Iterable[int]
        Data to write to bus.
    extended : bool, optional
        Use 64 bits offset independent of the platform, by default False.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    converted_buffer = (ViUInt16 * length)(*tuple(data))
    if extended:
        return library.viMoveOut16Ex(session, space, offset, length, converted_buffer)
    else:
        return library.viMoveOut16(session, space, offset, length, converted_buffer)


def move_out_32(library, session, space, offset, length, data, extended=False):
    """Moves an 32-bit block of data from local memory.

    Corresponds to viMoveOut32* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space into which move the data.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    length : int
        Number of elements to transfer, where the data width of
        the elements to transfer is identical to the source data width.
    data : Iterable[int]
        Data to write to bus.
    extended : bool, optional
        Use 64 bits offset independent of the platform, by default False.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.


    """
    converted_buffer = (ViUInt32 * length)(*tuple(data))
    if extended:
        return library.viMoveOut32Ex(session, space, offset, length, converted_buffer)
    else:
        return library.viMoveOut32(session, space, offset, length, converted_buffer)


def move_out_64(library, session, space, offset, length, data, extended=False):
    """Moves an 64-bit block of data from local memory.

    Corresponds to viMoveOut64* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space into which move the data.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    length : int
        Number of elements to transfer, where the data width of
        the elements to transfer is identical to the source data width.
    data : Iterable[int]
        Data to write to bus.
    extended : bool, optional
        Use 64 bits offset independent of the platform, by default False.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    converted_buffer = (ViUInt64 * length)(*tuple(data))
    if extended:
        return library.viMoveOut64Ex(session, space, offset, length, converted_buffer)
    else:
        return library.viMoveOut64(session, space, offset, length, converted_buffer)


# noinspection PyShadowingBuiltins
def open(
    library,
    session,
    resource_name,
    access_mode=constants.AccessModes.no_lock,
    open_timeout=constants.VI_TMO_IMMEDIATE,
):
    """Opens a session to the specified resource.

    Corresponds to viOpen function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISARMSession
        Resource Manager session (should always be a session returned from
        open_default_resource_manager()).
    resource_name : str
        Unique symbolic name of a resource.
    access_mode : constants.AccessModes, optional
        Specifies the mode by which the resource is to be accessed.
    open_timeout : int
        If the ``access_mode`` parameter requests a lock, then this
        parameter specifies the absolute time period (in milliseconds) that
        the resource waits to get unlocked before this operation returns an
        error.

    Returns
    -------
    VISASession
        Unique logical identifier reference to a session
    constants.StatusCode
        Return value of the library call.

    """
    try:
        open_timeout = int(open_timeout)
    except ValueError:
        raise ValueError(
            "open_timeout (%r) must be an integer (or compatible type)" % open_timeout
        )
    out_session = ViSession()

    # [ViSession, ViRsrc, ViAccessMode, ViUInt32, ViPSession]
    # ViRsrc converts from (str, unicode, bytes) to bytes
    ret = library.viOpen(
        session, resource_name, access_mode, open_timeout, byref(out_session)
    )
    return out_session.value, ret


def open_default_resource_manager(library):
    """This function returns a session to the Default Resource Manager resource.

    Corresponds to viOpenDefaultRM function of the VISA library.

    Returns
    -------
    VISARMSession
        Unique logical identifier to a Default Resource Manager session
    constants.StatusCode
        Return value of the library call.

    """
    session = ViSession()
    ret = library.viOpenDefaultRM(byref(session))
    return session.value, ret


def out_8(library, session, space, offset, data, extended=False):
    """Write an 8-bit value to the specified memory space and offset.

    Corresponds to viOut8* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space into which to write.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    data : int
        Data to write to bus.
    extended : bool, optional
        Use 64 bits offset independent of the platform.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    if extended:
        return library.viOut8Ex(session, space, offset, data)
    else:
        return library.viOut8(session, space, offset, data)


def out_16(library, session, space, offset, data, extended=False):
    """Write a 16-bit value to the specified memory space and offset.

    Corresponds to viOut16* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space into which to write.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    data : int
        Data to write to bus.
    extended : bool, optional
        Use 64 bits offset independent of the platform.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    if extended:
        return library.viOut16Ex(session, space, offset, data, extended=False)
    else:
        return library.viOut16(session, space, offset, data, extended=False)


def out_32(library, session, space, offset, data, extended=False):
    """Write a 32-bit value to the specified memory space and offset.

    Corresponds to viOut32* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space into which to write.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    data : int
        Data to write to bus.
    extended : bool, optional
        Use 64 bits offset independent of the platform.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    if extended:
        return library.viOut32Ex(session, space, offset, data)
    else:
        return library.viOut32(session, space, offset, data)


def out_64(library, session, space, offset, data, extended=False):
    """Write a 64-bit value to the specified memory space and offset.

    Corresponds to viOut64* functions of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    space : constants.AddressSpace
        Address space into which to write.
    offset : int
        Offset (in bytes) of the address or register from which to read.
    data : int
        Data to write to bus.
    extended : bool, optional
        Use 64 bits offset independent of the platform.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    if extended:
        return library.viOut64Ex(session, space, offset, data)
    else:
        return library.viOut64(session, space, offset, data)


def parse_resource(library, session, resource_name):
    """Parse a resource string to get the interface information.

    Corresponds to viParseRsrc function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISARMSession
        Resource Manager session (should always be the Default Resource
        Manager for VISA returned from open_default_resource_manager()).
    resource_name : str
            Unique symbolic name of a resource.

    Returns
    -------
    ResourceInfo
        Resource information with interface type and board number
    constants.StatusCode
        Return value of the library call.

    """
    interface_type = ViUInt16()
    interface_board_number = ViUInt16()

    # [ViSession, ViRsrc, ViPUInt16, ViPUInt16]
    # ViRsrc converts from (str, unicode, bytes) to bytes
    ret = library.viParseRsrc(
        session, resource_name, byref(interface_type), byref(interface_board_number)
    )
    return (
        ResourceInfo(
            constants.InterfaceType(interface_type.value),
            interface_board_number.value,
            None,
            None,
            None,
        ),
        ret,
    )


def parse_resource_extended(library, session, resource_name):
    """Parse a resource string to get extended interface information.

    Corresponds to viParseRsrcEx function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISARMSession
        Resource Manager session (should always be the Default Resource
        Manager for VISA returned from open_default_resource_manager()).
    resource_name : str
            Unique symbolic name of a resource.

    Returns
    -------
    ResourceInfo
        Resource information with interface type and board number
    constants.StatusCode
        Return value of the library call.

    """
    interface_type = ViUInt16()
    interface_board_number = ViUInt16()
    resource_class = create_string_buffer(constants.VI_FIND_BUFLEN)
    unaliased_expanded_resource_name = create_string_buffer(constants.VI_FIND_BUFLEN)
    alias_if_exists = create_string_buffer(constants.VI_FIND_BUFLEN)

    # [ViSession, ViRsrc, ViPUInt16, ViPUInt16, ViAChar, ViAChar, ViAChar]
    # ViRsrc converts from (str, unicode, bytes) to bytes
    ret = library.viParseRsrcEx(
        session,
        resource_name,
        byref(interface_type),
        byref(interface_board_number),
        resource_class,
        unaliased_expanded_resource_name,
        alias_if_exists,
    )

    res = [
        buffer_to_text(val)
        for val in (resource_class, unaliased_expanded_resource_name, alias_if_exists)
    ]

    if res[-1] == "":
        res[-1] = None

    return (
        ResourceInfo(
            constants.InterfaceType(interface_type.value),
            interface_board_number.value,
            *res,
        ),
        ret,
    )


def peek_8(library, session, address):
    """Read an 8-bit value from the specified address.

    Corresponds to viPeek8 function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    address : VISAMemoryAddress
        Source address to read the value.

    Returns
    -------
    int
        Data read from bus
    constants.StatusCode
        Return value of the library call.

    """
    value_8 = ViUInt8()
    ret = library.viPeek8(session, address, byref(value_8))
    return value_8.value, ret


def peek_16(library, session, address):
    """Read an 16-bit value from the specified address.

    Corresponds to viPeek16 function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    address : VISAMemoryAddress
        Source address to read the value.

    Returns
    -------
    int
        Data read from bus
    constants.StatusCode
        Return value of the library call.

    """
    value_16 = ViUInt16()
    ret = library.viPeek16(session, address, byref(value_16))
    return value_16.value, ret


def peek_32(library, session, address):
    """Read an 32-bit value from the specified address.

    Corresponds to viPeek32 function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    address : VISAMemoryAddress
        Source address to read the value.

    Returns
    -------
    int
        Data read from bus
    constants.StatusCode
        Return value of the library call.

    """
    value_32 = ViUInt32()
    ret = library.viPeek32(session, address, byref(value_32))
    return value_32.value, ret


def peek_64(library, session, address):
    """Read an 64-bit value from the specified address.

    Corresponds to viPeek64 function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    address : VISAMemoryAddress
        Source address to read the value.

    Returns
    -------
    int
        Data read from bus
    constants.StatusCode
        Return value of the library call.

    """
    value_64 = ViUInt64()
    ret = library.viPeek64(session, address, byref(value_64))
    return value_64.value, ret


def poke_8(library, session, address, data):
    """Write an 8-bit value to the specified address.

    Corresponds to viPoke8 function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    address : VISAMemoryAddress
        Source address to read the value.
    data : int
        Data to write.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viPoke8(session, address, data)


def poke_16(library, session, address, data):
    """Write an 16-bit value to the specified address.

    Corresponds to viPoke16 function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    address : VISAMemoryAddress
        Source address to read the value.
    data : int
        Data to write.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viPoke16(session, address, data)


def poke_32(library, session, address, data):
    """Write an 32-bit value to the specified address.

    Corresponds to viPoke32 function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    address : VISAMemoryAddress
        Source address to read the value.
    data : int
        Data to write.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viPoke32(session, address, data)


def poke_64(library, session, address, data):
    """Write an 64-bit value to the specified address.

    Corresponds to viPoke64 function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    address : VISAMemoryAddress
        Source address to read the value.
    data : int
        Data to write.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viPoke64(session, address, data)


def read(library, session, count):
    """Reads data from device or interface synchronously.

    Corresponds to viRead function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    count : int
        Number of bytes to be read.

    Returns
    -------
    bytes
        Date read
    constants.StatusCode
        Return value of the library call.

    """
    buffer = create_string_buffer(count)
    return_count = ViUInt32()
    ret = library.viRead(session, buffer, count, byref(return_count))
    return buffer.raw[: return_count.value], ret


def read_stb(library, session):
    """Reads a status byte of the service request.

    Corresponds to viReadSTB function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.

    Returns
    -------
    int
        Service request status byte
    constants.StatusCode
        Return value of the library call.

    """
    status = ViUInt16()
    ret = library.viReadSTB(session, byref(status))
    return status.value, ret


def read_to_file(library, session, filename, count):
    """Read data synchronously, and store the transferred data in a file.

    Corresponds to viReadToFile function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    filename : str
        Name of file to which data will be written.
    count : int
        Number of bytes to be read.

    Returns
    -------
    int
        Number of bytes actually transferred
    constants.StatusCode
        Return value of the library call.

    """
    return_count = ViUInt32()
    ret = library.viReadToFile(session, filename, count, return_count)
    return return_count, ret


def set_attribute(library, session, attribute, attribute_state):
    """Set the state of an attribute.

    Corresponds to viSetAttribute function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    attribute : constants.ResourceAttribute
        Attribute for which the state is to be modified.
    attribute_state : Any
        The state of the attribute to be set for the specified object.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viSetAttribute(session, attribute, attribute_state)


def set_buffer(library, session, mask, size):
    """Set the size for the formatted I/O and/or low-level I/O communication buffer(s).

    Corresponds to viSetBuf function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    mask : constants.BufferType
        Specifies the type of buffer.
    size : int
        The size to be set for the specified buffer(s).

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viSetBuf(session, mask, size)


def status_description(library, session, status):
    """Return a user-readable description of the status code passed to the operation.

    Corresponds to viStatusDesc function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    status : constants.StatusCode
        Status code to interpret.

    Returns
    -------
    str
        User-readable string interpretation of the status code.
    constants.StatusCode
        Return value of the library call.

    """
    description = create_string_buffer(256)
    ret = library.viStatusDesc(session, status, description)
    return buffer_to_text(description), ret


def terminate(library, session, degree, job_id):
    """Request a VISA session to terminate normal execution of an operation.

    Corresponds to viTerminate function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    degree : None
        Not used in this version of the VISA specification.
    job_id : VISAJobId
        Specifies an operation identifier. If a user passes None as the
        job_id value to viTerminate(), a VISA implementation should abort
        any calls in the current process executing on the specified vi.
        Any call that is terminated this way should return VI_ERROR_ABORT.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viTerminate(session, degree, job_id)


def uninstall_handler(library, session, event_type, handler, user_handle=None):
    """Uninstall handlers for events.

    Corresponds to viUninstallHandler function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    event_type : constants.EventType
        Logical event identifier.
    handler : VISAHandler
        Handler to be uninstalled by a client application.
    user_handle:
        A value specified by an application that can be used for
        identifying handlers uniquely in a session for an event.
        The modified value of the user_handle as returned by install_handler
        should be used instead of the original value.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    with set_user_handle_type(library, user_handle):
        if user_handle is not None:
            user_handle = byref(user_handle)
        return library.viUninstallHandler(session, event_type, handler, user_handle)


def unlock(library, session):
    """Relinquish a lock for the specified resource.

    Corresponds to viUnlock function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viUnlock(session)


def unmap_address(library, session):
    """Unmap memory space previously mapped by map_address().

    Corresponds to viUnmapAddress function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viUnmapAddress(session)


def unmap_trigger(library, session, trigger_source, trigger_destination):
    """Undo a previous map between a trigger source line and a destination line.

    Corresponds to viUnmapTrigger function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    trigger_source : constants.InputTriggerLine
        Source line used in previous map.
    trigger_destination : constants.OutputTriggerLine
        Destination line used in previous map.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    return library.viUnmapTrigger(session, trigger_source, trigger_destination)


def usb_control_in(
    library,
    session,
    request_type_bitmap_field,
    request_id,
    request_value,
    index,
    length=0,
):
    """Perform a USB control pipe transfer from the device.

    Corresponds to viUsbControlIn function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    request_type_bitmap_field : int
        bmRequestType parameter of the setup stage of a USB control transfer.
    request_id : int
        bRequest parameter of the setup stage of a USB control transfer.
    request_value : int
        wValue parameter of the setup stage of a USB control transfer.
    index : int
        wIndex parameter of the setup stage of a USB control transfer.
        This is usually the index of the interface or endpoint.
    length : int, optional
        wLength parameter of the setup stage of a USB control transfer.
        This value also specifies the size of the data buffer to receive
        the data from the optional data stage of the control transfer.

    Returns
    -------
    bytes
        The data buffer that receives the data from the optional data stage
        of the control transfer
    constants.StatusCode
        Return value of the library call.

    """
    buffer = create_string_buffer(length)
    return_count = ViUInt16()
    ret = library.viUsbControlIn(
        session,
        request_type_bitmap_field,
        request_id,
        request_value,
        index,
        length,
        buffer,
        byref(return_count),
    )
    return buffer.raw[: return_count.value], ret


def usb_control_out(
    library,
    session,
    request_type_bitmap_field,
    request_id,
    request_value,
    index,
    data="",
):
    """Perform a USB control pipe transfer to the device.

    Corresponds to viUsbControlOut function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    request_type_bitmap_field : int
        bmRequestType parameter of the setup stage of a USB control transfer.
    request_id : int
        bRequest parameter of the setup stage of a USB control transfer.
    request_value : int
        wValue parameter of the setup stage of a USB control transfer.
    index : int
        wIndex parameter of the setup stage of a USB control transfer.
        This is usually the index of the interface or endpoint.
    data : bytes, optional
        The data buffer that sends the data in the optional data stage of
        the control transfer.

    Returns
    -------
    constants.StatusCode
        Return value of the library call.

    """
    length = len(data)
    return library.viUsbControlOut(
        session,
        request_type_bitmap_field,
        request_id,
        request_value,
        index,
        length,
        data,
    )


def vxi_command_query(library, session, mode, command):
    """Send the device a miscellaneous command or query and/or retrieves the response to a previous query.

    Corresponds to viVxiCommandQuery function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    mode : constants.VXICommands
        Specifies whether to issue a command and/or retrieve a response.
    command : int
        The miscellaneous command to send.

    Returns
    -------
    int
        The response retrieved from the device
    constants.StatusCode
        Return value of the library call.

    """
    response = ViUInt32()
    ret = library.viVxiCommandQuery(session, mode, command, byref(response))
    return response.value, ret


def wait_on_event(library, session, in_event_type, timeout):
    """Wait for an occurrence of the specified event for a given session.

    Corresponds to viWaitOnEvent function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    in_event_type : constants.EventType
        Logical identifier of the event(s) to wait for.
    timeout : int
        Absolute time period in time units that the resource shall wait for
        a specified event to occur before returning the time elapsed error.
        The time unit is in milliseconds.

    Returns
    -------
    constants.EventType
        Logical identifier of the event actually received
    VISAEventContext
        A handle specifying the unique occurrence of an event
    constants.StatusCode
        Return value of the library call.

    """
    out_event_type = ViEventType()
    out_context = ViEvent()
    ret = library.viWaitOnEvent(
        session, in_event_type, timeout, byref(out_event_type), byref(out_context)
    )
    return out_event_type.value, out_context, ret


def write(library, session, data):
    """Write data to device or interface synchronously.

    Corresponds to viWrite function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    data : bytes
        Data to be written.

    Returns
    -------
    int
        Number of bytes actually transferred
    constants.StatusCode
        Return value of the library call.

    """
    return_count = ViUInt32()
    # [ViSession, ViBuf, ViUInt32, ViPUInt32]
    ret = library.viWrite(session, data, len(data), byref(return_count))
    return return_count.value, ret


def write_asynchronously(library, session, data):
    """Write data to device or interface asynchronously.

    Corresponds to viWriteAsync function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    data : bytes
        Data to be written.

    Returns
    -------
    VISAJobID
        Job ID of this asynchronous write operation
    constants.StatusCode
        Return value of the library call.

    """
    job_id = ViJobId()
    # [ViSession, ViBuf, ViUInt32, ViPJobId]
    ret = library.viWriteAsync(session, data, len(data), byref(job_id))
    return job_id, ret


def write_from_file(library, session, filename, count):
    """Take data from a file and write it out synchronously.

    Corresponds to viWriteFromFile function of the VISA library.

    Parameters
    ----------
    library : ctypes.WinDLL or ctypes.CDLL
        ctypes wrapped library.
    session : VISASession
        Unique logical identifier to a session.
    filename : str
        Name of file from which data will be read.
    count : int
        Number of bytes to be written.

    Returns
    -------
    int
        Number of bytes actually transferred
    constants.StatusCode
        Return value of the library call.

    """
    return_count = ViUInt32()
    ret = library.viWriteFromFile(session, filename, count, return_count)
    return return_count, ret



# -*- coding: utf-8 -*-
"""Highlevel wrapper of the VISA Library.

This file is part of PyVISA.

:copyright: 2014-2022 by PyVISA Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""
import ctypes
import logging
from collections import OrderedDict
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    SupportsBytes,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from pyvisa import constants, errors, highlevel, logger, typing

from ..util import (
    DebugInfo,
    LibraryPath,
    add_user_dll_extra_paths,
    read_user_library_path,
)
from . import functions, types
from .cthelper import Library, find_library

logger = logging.LoggerAdapter(logger, {"backend": "ivi"})  # type: ignore


T = TypeVar("T", bound=highlevel.VisaLibraryBase)


def add_visa_methods(aclass: Type[T]) -> Type[T]:
    """Add the methods to the library."""
    for method in functions.visa_functions:
        setattr(aclass, method, getattr(functions, method))
    return aclass


def _args_to_str(args: Tuple[Any, ...]) -> Tuple[str, ...]:
    """Convert function arguments to str."""
    out = []
    for arg in args:
        try:
            out.append(str(arg._obj))
        except AttributeError:
            out.append(arg)
    return tuple(out)


def unique(seq):
    """Keep unique while preserving order."""
    seen = set()
    return [x for x in seq if not (x in seen or seen.add(x))]


@add_visa_methods
class IVIVisaLibrary(highlevel.VisaLibraryBase):
    """High level IVI-VISA Library wrapper using ctypes.

    The easiest way to instantiate the library is to let `pyvisa` find the
    right one for you. This looks first in your configuration file (~/.pyvisarc).
    If it fails, it uses `ctypes.util.find_library` to try to locate a library
    in a way similar to what the compiler does:

       >>> visa_library = IVIVisaLibrary()

    But you can also specify the path:

        >>> visa_library = IVIVisaLibrary('/my/path/visa.so')

    """

    @staticmethod
    def get_library_paths() -> Tuple[LibraryPath, ...]:
        """Return a tuple of possible library paths."""
        # Add extra .dll dependency search paths before attempting to load libraries
        add_user_dll_extra_paths()

        # Try to find IVI libraries using known names.
        tmp: List[str]
        tmp = [
            find_library(library_path)
            for library_path in ("visa", "visa32", "visa32.dll", "visa64", "visa64.dll")
        ]

        logger.debug("Automatically found library files: %s" % tmp)

        # Prepend the path provided by the user in configuration files (if any).
        user_lib = read_user_library_path()
        if user_lib:
            tmp.insert(0, user_lib)

        # Deduplicate and convert string paths to LibraryPath objects
        return tuple(
            LibraryPath(library_path)
            for library_path in unique(tmp)
            if library_path is not None
        )

    @classmethod
    def get_debug_info(cls) -> DebugInfo:
        """Return a list of lines with backend info."""
        from pyvisa import __version__

        d: Dict[str, Any] = OrderedDict()
        d["Version"] = "%s (bundled with PyVISA)" % __version__

        paths = cls.get_library_paths()

        for ndx, visalib in enumerate(paths, 1):
            nfo: Dict[str, Union[str, List[str]]] = OrderedDict()
            nfo["found by"] = visalib.found_by
            nfo["architecture"] = [str(a.value) for a in visalib.arch]
            try:
                lib = cls(visalib)
                sess, _ = lib.open_default_resource_manager()
                nfo["Vendor"] = str(
                    lib.get_attribute(
                        sess, constants.ResourceAttribute.resource_manufacturer_name
                    )[0]
                )
                nfo["Impl. Version"] = str(
                    lib.get_attribute(
                        sess, constants.ResourceAttribute.resource_impl_version
                    )[0]
                )
                nfo["Spec. Version"] = str(
                    lib.get_attribute(
                        sess, constants.ResourceAttribute.resource_spec_version
                    )[0]
                )
                lib.close(sess)
            except Exception as e:
                err_string = str(e)
                if "No matching architecture" in err_string:
                    nfo[
                        "Could not get more info"
                    ] = "Interpreter and library have different bitness."
                else:
                    nfo["Could not get more info"] = err_string.split("\n")

            d["#%d: %s" % (ndx, visalib)] = nfo

        if not paths:
            d["Binary library"] = "Not found"

        return d

    def _init(self) -> None:
        try:
            lib = Library(self.library_path)
        except OSError as exc:
            raise errors.LibraryError.from_exception(exc, self.library_path)

        self.lib = lib
        self._async_read_jobs: List[Tuple[types.ViJobId, SupportsBytes]] = []

        # Set the argtypes, restype and errcheck for each function
        # of the visa library. Additionally store in `_functions` the
        # name of the functions.
        functions.set_signatures(self.lib, errcheck=self._return_handler)

        logger.debug(
            "Library signatures: %d ok, %d failed",
            len(getattr(self.lib, "_functions", [])),
            len(getattr(self.lib, "_functions_failed", [])),
        )

        # Set the library functions as attributes of the object.
        for method_name in getattr(self.lib, "_functions", []):
            setattr(self, method_name, getattr(self.lib, method_name))

    def _return_handler(self, ret_value: int, func: Callable, arguments: tuple) -> Any:
        """Check return values for errors and warnings."""
        logger.debug(
            "%s%s -> %r",
            func.__name__,
            _args_to_str(arguments),
            ret_value,
            extra=self._logging_extra,
        )

        # The first argument of almost all registered visa functions is a session.
        # We store the error code per session
        session = None
        if func.__name__ not in ("viFindNext",):
            try:
                session = arguments[0]
            except KeyError:
                raise Exception(
                    "Function %r does not seem to be a valid "
                    "visa function (len args %d)" % (func, len(arguments))
                )

            # Functions that use the first parameter to get a session value.
            if func.__name__ in ("viOpenDefaultRM",):
                session = session._obj.value

            if not isinstance(session, int):
                # Functions that might or might not have a session in the first argument.
                if func.__name__ not in (
                    "viClose",
                    "viGetAttribute",
                    "viSetAttribute",
                    "viStatusDesc",
                ):
                    raise Exception(
                        "Function %r does not seem to be a valid "
                        "visa function (type args[0] %r)" % (func, type(session))
                    )

                # Set session back to a safe value
                session = None

        return self.handle_return_value(session, ret_value)  # type: ignore

    def list_resources(
        self, session: typing.VISARMSession, query: str = "?*::INSTR"
    ) -> Tuple[str, ...]:
        r"""Returns a tuple of all connected devices matching query.

        Parameters
        ----------
        session : typing.VISARMSession
            Unused, present for consistency
        query : str
            Query using the VISA Resource Regular Expression syntax - which is
            not the same as the Python regular expression syntax. (see below)

        Notes
        -----

        The VISA Resource Regular Expression syntax is defined in the VISA
        Library specification:
        http://www.ivifoundation.org/docs/vpp43.pdf

        Symbol      Meaning
        ----------  ----------

        ?           Matches any one character.

        \           Makes the character that follows it an ordinary character
                    instead of special character. For example, when a question
                    mark follows a backslash (\?), it matches the ? character
                    instead of any one character.

        [list]      Matches any one character from the enclosed list. You can
                    use a hyphen to match a range of characters.

        [^list]     Matches any character not in the enclosed list. You can use
                    a hyphen to match a range of characters.

        *           Matches 0 or more occurrences of the preceding character or
                    expression.

        +           Matches 1 or more occurrences of the preceding character or
                    expression.

        Exp|exp     Matches either the preceding or following expression. The or
                    operator | matches the entire expression that precedes or
                    follows it and not just the character that precedes or follows
                    it. For example, VXI|GPIB means (VXI)|(GPIB), not VX(I|G)PIB.

        (exp)       Grouping characters or expressions.

        Thus the default query, '?*::INSTR', matches any sequences of characters
        ending ending with '::INSTR'.

        """
        resources: List[str] = []

        # Ignore some type checks because method are dynamically set
        try:
            (
                find_list,
                return_counter,
                instrument_description,
                err,
            ) = self._find_resources(  # type: ignore
                session, query
            )
        except errors.VisaIOError as e:
            if e.error_code == constants.StatusCode.error_resource_not_found:
                return tuple()
            raise e

        try:
            resources.append(instrument_description)
            for i in range(return_counter - 1):
                resources.append(self._find_next(find_list)[0])  # type: ignore
        finally:
            # This is the only occurrence of a find list so ignore the typing error
            # since it would make things harder to follow to document ViFindList as
            # supported.
            self.close(find_list)  # type: ignore

        return tuple(resource for resource in resources)

    def read_asynchronously(
        self, session: typing.VISASession, count: int
    ) -> Tuple[SupportsBytes, types.ViJobId, constants.StatusCode]:
        """Read data from device or interface asynchronously.

        Corresponds to viReadAsync function of the VISA library. Since the
        asynchronous operation may complete before the function call return
        implementation should make sure that get_buffer_from_id will be able
        to return the proper buffer before this method returns.

        Parameters
        ----------
        session : typing.VISASession
            Unique logical identifier to a session.
        count : int
            Number of bytes to be read.

        Returns
        -------
        SupportsBytes
            Buffer in which the data will be written.
        types.ViJobId
            Id of the job.
        constants.StatusCode
            Return value of the library call.

        """
        # The buffer actually supports bytes but typing fails
        buffer = ctypes.create_string_buffer(count)
        job_id = types.ViJobId()
        self._async_read_jobs.append((job_id, buffer))  # type: ignore
        ret = self.lib.viReadAsync(session, buffer, count, ctypes.byref(job_id))
        return buffer, job_id.value, ret  # type: ignore

    def get_buffer_from_id(self, job_id: typing.VISAJobID) -> Optional[SupportsBytes]:
        """Retrieve the buffer associated with a job id created in read_asynchronously.

        Parameters
        ----------
        job_id : VISAJobID
            Id of the job for which to retrieve the buffer.

        Returns
        -------
        SupportsBytes
            Buffer in which the data are stored.

        """
        for jid, buffer in self._async_read_jobs:
            if job_id == jid.value:
                return buffer

        return None



# -*- coding: utf-8 -*-
"""Test the behavior of the command line tools.

"""
import argparse
import sys
from subprocess import PIPE, Popen, run

import pytest

from pyvisa import cmd_line_tools, util

from . import BaseTestCase, require_visa_lib


class TestCmdLineTools(BaseTestCase):
    """Test the cmd line tools functions and scripts."""

    def test_visa_main_argument_handling(self):
        """Test we reject invalid values in visa_main."""
        from pyvisa.cmd_line_tools import visa_main

        old = sys.argv = ["python"]
        try:
            with pytest.raises(ValueError):
                visa_main("unknown")
        finally:
            sys.argv = old

    def test_visa_info(self):
        """Test the visa info command line tool."""
        result = run("pyvisa-info", stdout=PIPE, universal_newlines=True)
        details = util.system_details_to_str(util.get_system_details())
        # Path difference can lead to subtle differences in the backends
        # and Python path so compare only the first lines.
        assert result.stdout.strip().split("\n")[:5] == details.strip().split("\n")[:5]

    # TODO test backend selection: this is not easy at all to assert
    @require_visa_lib
    def test_visa_shell(self):
        """Test the visa shell function."""
        with Popen(["pyvisa-shell"], stdin=PIPE, stdout=PIPE) as p:
            stdout, stderr = p.communicate(b"exit")
        assert stdout.count(b"Welcome to the VISA shell") == 1


@pytest.mark.parametrize(
    "args, want",
    [
        (argparse.Namespace(backend=None), ""),
        (argparse.Namespace(backend="py"), "@py"),
        (argparse.Namespace(backend="foo.yaml@sim"), "foo.yaml@sim"),
        (argparse.Namespace(backend="/foo/bar/baz.yaml@sim"), "/foo/bar/baz.yaml@sim"),
        (argparse.Namespace(backend="@sim"), "@sim"),
    ],
)
def test__create_backend_str(args: argparse.Namespace, want: str) -> None:
    got = cmd_line_tools._create_backend_str(args)
    assert got == want



# -*- coding: utf-8 -*-
"""Test the handling of errors.

"""
import pickle

from pyvisa import errors
from pyvisa.testsuite import BaseTestCase


class TestPicleUnpickle(BaseTestCase):
    def _test_pickle_unpickle(self, instance):
        pickled = pickle.dumps(instance)
        unpickled = pickle.loads(pickled)
        assert isinstance(unpickled, type(instance))
        for attr in instance.__dict__:
            assert getattr(instance, attr) == getattr(unpickled, attr)

    def test_VisaIOError(self):
        self._test_pickle_unpickle(errors.VisaIOError(0))

    def test_VisaIOWarning(self):
        self._test_pickle_unpickle(errors.VisaIOWarning(0))

    def test_UnknownHandler(self):
        self._test_pickle_unpickle(errors.UnknownHandler(0, 0, 0))

    def test_OSNotSupported(self):
        self._test_pickle_unpickle(errors.OSNotSupported(""))

    def test_InvalidBinaryFormat(self):
        self._test_pickle_unpickle(errors.InvalidBinaryFormat())
        self._test_pickle_unpickle(errors.InvalidBinaryFormat("test"))

    def test_InvalidSession(self):
        self._test_pickle_unpickle(errors.InvalidSession())


class TestLibraryError(BaseTestCase):
    """Test the creation of Library errors."""

    def test_from_exception_not_found(self):
        """Test handling a missing library file."""
        exc = errors.LibraryError.from_exception(
            ValueError("visa.dll: image not found"), "visa.dll"
        )
        assert "File not found" in str(exc)

    def test_from_exception_wrong_arch(self):
        """Test handling a library that report the wrong bitness."""
        exc = errors.LibraryError.from_exception(
            ValueError("visa.dll: no suitable image found. no matching architecture"),
            "visa.dll",
        )
        assert "No matching architecture" in str(exc)

    def test_from_exception_wrong_filetype(self):
        """Test handling a library file of the wrong type."""
        exc = errors.LibraryError.from_exception(
            ValueError("visa.dll: no suitable image found."), "visa.dll"
        )
        assert "Could not determine filetype" in str(exc)

    def test_from_exception_wrong_ELF(self):
        """Test handling a library file with a wrong ELF."""
        exc = errors.LibraryError.from_exception(
            ValueError("visa.dll: wrong ELF class"), "visa.dll"
        )
        assert "No matching architecture" in str(exc)

    def test_from_exception_random(self):
        """Test handling a library for which the error is not a usual one."""
        exc = errors.LibraryError.from_exception(ValueError("visa.dll"), "visa.dll")
        assert "Error while accessing" in str(exc)

    def test_from_exception_decode_error(self):
        """Test handling an error that decode to string."""

        class DummyExc(Exception):
            def __str__(self):
                raise b"\xFF".decode("ascii")

        exc = errors.LibraryError.from_exception(
            DummyExc("visa.dll: wrong ELF class"), "visa.dll"
        )
        assert "Error while accessing visa.dll." == str(exc)

    # from_wrong_arch is exercised through the above tests.



# -*- coding: utf-8 -*-
"""Test the capabilities of the ResourceManager.

"""
import gc
import logging
import re

import pytest

from pyvisa import InvalidSession, ResourceManager, VisaIOError, errors
from pyvisa.constants import AccessModes, InterfaceType, StatusCode
from pyvisa.highlevel import VisaLibraryBase
from pyvisa.rname import ResourceName
from pyvisa.testsuite import BaseTestCase

from . import RESOURCE_ADDRESSES, require_virtual_instr


@require_virtual_instr
class TestResourceManager:
    """Test the pyvisa ResourceManager."""

    def setup_method(self):
        """Create a ResourceManager with the default backend library."""
        self.rm = ResourceManager()

    def teardown_method(self):
        """Close the ResourceManager."""
        if self.rm is not None:
            self.rm.close()
            del self.rm
            gc.collect()

    def test_lifecycle(self, caplog):
        """Test creation and closing of the resource manager."""
        assert self.rm.session is not None
        assert self.rm.visalib is not None
        assert self.rm is self.rm.visalib.resource_manager
        assert not self.rm.list_opened_resources()

        assert self.rm.visalib is ResourceManager(self.rm.visalib).visalib

        with caplog.at_level(level=logging.DEBUG, logger="pyvisa"):
            self.rm.close()

        assert caplog.records

        with pytest.raises(InvalidSession):
            self.rm.session
        assert self.rm.visalib.resource_manager is None

    # This test is flaky
    @pytest.mark.skip
    def test_cleanup_on_del(self, caplog):
        """Test that deleting the rm does clean the VISA session"""
        # The test seems to assert what it should even though the coverage report
        # seems wrong
        rm = self.rm
        self.rm = None
        with caplog.at_level(logging.DEBUG, logger="pyvisa"):
            del rm
            gc.collect()

        assert "Closing ResourceManager" in caplog.records[0].message

    def test_resource_manager_unicity(self):
        """Test the resource manager is unique per backend as expected."""
        new_rm = ResourceManager()
        assert self.rm is new_rm
        assert self.rm.session == new_rm.session

    def test_str(self):
        """Test computing the string representation of the resource manager"""
        assert re.match(r"Resource Manager of .*", str(self.rm))
        self.rm.close()
        assert re.match(r"Resource Manager of .*", str(self.rm))

    def test_repr(self):
        """Test computing the repr of the resource manager"""
        assert re.match(r"<ResourceManager\(<.*>\)>", repr(self.rm))
        self.rm.close()
        assert re.match(r"<ResourceManager\(<.*>\)>", repr(self.rm))

    def test_last_status(self):
        """Test accessing the status of the last operation."""
        assert self.rm.last_status == StatusCode.success

        # Access the generic last status through the visalib
        assert self.rm.last_status == self.rm.visalib.last_status

        # Test accessing the status for an invalid session
        with pytest.raises(errors.Error) as cm:
            self.rm.visalib.get_last_status_in_session("_nonexisting_")
        assert "The session" in cm.exconly()

    def test_list_resource(self):
        """Test listing the available resources."""
        # Default settings
        resources = self.rm.list_resources()
        for v in (v for v in RESOURCE_ADDRESSES.values() if v.endswith("INSTR")):
            assert str(ResourceName.from_string(v)) in resources

        # All resources
        resources = self.rm.list_resources("?*")
        for v in RESOURCE_ADDRESSES.values():
            assert str(ResourceName.from_string(v)) in resources

    def test_accessing_resource_infos(self):
        """Test accessing resource infos."""
        rname = list(RESOURCE_ADDRESSES.values())[0]
        rinfo_ext = self.rm.resource_info(rname)
        rinfo = self.rm.resource_info(rname, extended=False)

        rname = ResourceName().from_string(rname)
        assert rinfo_ext.interface_type == getattr(
            InterfaceType, rname.interface_type.lower()
        )
        assert rinfo_ext.interface_board_number == int(rname.board)
        assert rinfo_ext.resource_class == rname.resource_class
        assert rinfo_ext.resource_name == str(rname)

        assert rinfo.interface_type == getattr(
            InterfaceType, rname.interface_type.lower()
        )
        assert rinfo.interface_board_number == int(rname.board)

    def test_listing_resource_infos(self):
        """Test listing resource infos."""
        infos = self.rm.list_resources_info()

        for rname, rinfo_ext in infos.items():
            rname = ResourceName().from_string(rname)
            assert rinfo_ext.interface_type == getattr(
                InterfaceType, rname.interface_type.lower()
            )
            assert rinfo_ext.interface_board_number == int(rname.board)
            assert rinfo_ext.resource_class == rname.resource_class
            assert rinfo_ext.resource_name == str(rname)

    def test_opening_resource(self):
        """Test opening and closing resources."""
        rname = list(RESOURCE_ADDRESSES.values())[0]
        rsc = self.rm.open_resource(rname, timeout=1234)

        # Check the resource is listed as opened and the attributes are right.
        assert rsc in self.rm.list_opened_resources()
        assert rsc.timeout == 1234

        # Close the rm to check that we close all resources.
        self.rm.close()

        assert not self.rm.list_opened_resources()
        with pytest.raises(InvalidSession):
            rsc.session

    def test_opening_resource_bad_open_timeout(self):
        """Test opening a resource with a non integer open_timeout."""
        rname = list(RESOURCE_ADDRESSES.values())[0]

        with pytest.raises(ValueError) as cm:
            self.rm.open_resource(rname, open_timeout="")

        assert "integer (or compatible type)" in str(cm.exconly())

    def test_opening_resource_with_lock(self):
        """Test opening a locked resource"""
        rname = list(RESOURCE_ADDRESSES.values())[0]
        rsc = self.rm.open_resource(rname, access_mode=AccessModes.exclusive_lock)
        assert len(self.rm.list_opened_resources()) == 1

        # Timeout when accessing a locked resource
        with pytest.raises(VisaIOError):
            self.rm.open_resource(rname, access_mode=AccessModes.exclusive_lock)
        assert len(self.rm.list_opened_resources()) == 1

        # Success to access an unlocked resource.
        rsc.unlock()
        with self.rm.open_resource(
            rname, access_mode=AccessModes.exclusive_lock
        ) as rsc2:
            assert rsc.session != rsc2.session
            assert len(self.rm.list_opened_resources()) == 2

    def test_opening_resource_specific_class(self):
        """Test opening a resource requesting a specific class."""
        rname = list(RESOURCE_ADDRESSES.values())[0]
        with pytest.raises(TypeError):
            self.rm.open_resource(rname, resource_pyclass=object)

        assert len(self.rm.list_opened_resources()) == 0

    def test_open_resource_unknown_resource_type(self, caplog):
        """Test opening a resource for which no registered class exist."""
        rc = ResourceManager._resource_classes
        old = rc.copy()

        class FakeResource:
            def __init__(self, *args):
                raise RuntimeError()

        rc[(InterfaceType.unknown, "")] = FakeResource
        del rc[(InterfaceType.tcpip, "INSTR")]

        rm = ResourceManager()
        try:
            caplog.clear()
            with caplog.at_level(level=logging.DEBUG, logger="pyvisa"):
                with pytest.raises(RuntimeError):
                    rm.open_resource("TCPIP::192.168.0.1::INSTR")
            assert caplog.records
        finally:
            ResourceManager._resource_classes = old

    def test_opening_resource_unknown_attribute(self):
        """Test opening a resource and attempting to set an unknown attr."""
        rname = list(RESOURCE_ADDRESSES.values())[0]
        with pytest.raises(ValueError):
            self.rm.open_resource(rname, unknown_attribute=None)

        assert len(self.rm.list_opened_resources()) == 0


@require_virtual_instr
class TestResourceParsing(BaseTestCase):
    """Test parsing resources using the builtin mechanism and the VISA lib.

    Those tests require that the interface exist (at least in Keysight
    implementation) so we cannot test arbitrary interfaces (PXI for example).

    """

    def setup_method(self):
        """Create a ResourceManager with the default backend library."""
        self.rm = ResourceManager()

    def teardown_method(self):
        """Close the ResourceManager."""
        del self.rm
        gc.collect()

    def test_parse_tcpip_instr(self):
        self._parse_test("TCPIP::192.168.200.200::INSTR")

    def test_parse_tcpip_socket(self):
        self._parse_test("TCPIP::192.168.200.200::7020::SOCKET")

    def _parse_test(self, rn):
        # Visa lib
        p = self.rm.visalib.parse_resource_extended(self.rm.session, rn)

        # Internal
        pb = VisaLibraryBase.parse_resource_extended(
            self.rm.visalib, self.rm.session, rn
        )
        assert p == pb

        # Non-extended parsing
        # Visa lib
        p = self.rm.visalib.parse_resource(self.rm.session, rn)

        # Internal
        pb = VisaLibraryBase.parse_resource(self.rm.visalib, self.rm.session, rn)
        assert p == pb



# -*- coding: utf-8 -*-
"""

"""
from pyvisa import constants
from pyvisa.highlevel import VisaLibraryBase
from pyvisa.util import LibraryPath


class FakeResource:
    def close(self):
        pass


class FalseVISALib(VisaLibraryBase):
    pass

    @staticmethod
    def get_library_paths():
        return (LibraryPath("unset"),)

    def _init(self):
        pass

    def open_resource(self, *args, **kwargs):
        self.open_resource_called = True
        return FakeResource()

    def open_default_resource_manager(self):
        return 1, constants.StatusCode.success

    def close(self, session):
        pass


WRAPPER_CLASS = FalseVISALib



# -*- coding: utf-8 -*-
"""Common test case for all resources.

"""
import gc
import logging
import re
from typing import Union

import pytest

from pyvisa import InvalidSession, ResourceManager
from pyvisa.constants import InterfaceType, ResourceAttribute, StatusCode, Timeouts
from pyvisa.resources.resource import Resource
from pyvisa.rname import ResourceName

from . import RESOURCE_ADDRESSES


class ResourceTestCase:
    """Base test case for all resources."""

    #: Type of resource being tested in this test case.
    #: See RESOURCE_ADDRESSES in the __init__.py file of this package for
    #: acceptable values
    RESOURCE_TYPE = ""

    #: Minimal timeout value accepted by the resource. When setting the timeout
    #: to Timeouts.immediate, Visa (Keysight at least) may actually use a
    #: different value depending on the values supported by the resource.
    MINIMAL_TIMEOUT: Union[int, Timeouts] = Timeouts.immediate

    def setup_method(self):
        """Create a resource using the address matching the type."""
        name = RESOURCE_ADDRESSES[self.RESOURCE_TYPE]
        self.rname = ResourceName.from_string(name)
        self.rm = ResourceManager()
        self.instr = self.rm.open_resource(name)
        self.instr.clear()

    def teardown_method(self):
        """Close the resource at the end of the test."""
        if self.instr:
            self.instr.close()
        if self.rm:
            self.rm.close()

    def test_lifecycle(self):
        """Test the lifecyle of a resource and the use as a context manager."""
        assert self.instr.session is not None
        assert self.instr.visalib is not None
        assert self.instr.last_status == StatusCode.success

        self.instr.close()

        with pytest.raises(InvalidSession):
            self.instr.session

        with self.rm.open_resource(str(self.rname), read_termination="\0") as instr:
            assert len(self.rm.list_opened_resources()) == 1
            assert instr.read_termination == "\0"
        assert len(self.rm.list_opened_resources()) == 0

    def test_close_on_del(self, caplog):
        """Test the lifecyle of a resource and the use as a context manager."""
        with caplog.at_level(logging.DEBUG):
            self.instr = None
            gc.collect()

        assert "- closing" in caplog.records[0].message
        assert "- is closed", caplog.output[-1].message

    def test_alias_bypassing(self):
        """Test that a resource that cannot normalize an alias keep the alias."""
        instr = Resource(self.rm, "visa_alias")
        assert re.match(r".* at %s" % "visa_alias", str(instr))

    def test_str(self):
        """Test the string representation of a resource."""
        assert re.match(r".* at %s" % str(self.rname), str(self.instr))
        self.instr.close()
        assert re.match(r".* at %s" % str(self.rname), str(self.instr))

    def test_repr(self):
        """Test the repr of a resource."""
        assert re.match(r"<.*\('%s'\)>" % str(self.rname), repr(self.instr))
        self.instr.close()
        assert re.match(r"<.*\('%s'\)>" % str(self.rname), repr(self.instr))

    def test_timeout(self):
        """Test setting the timeout attribute."""
        self.instr.timeout = None
        assert self.instr.timeout == float("+inf")
        assert (
            self.instr.get_visa_attribute(ResourceAttribute.timeout_value)
            == Timeouts.infinite
        )

        self.instr.timeout = 0.1
        assert self.instr.timeout == 1
        assert (
            self.instr.get_visa_attribute(ResourceAttribute.timeout_value)
            == self.MINIMAL_TIMEOUT
        )

        self.instr.timeout = 10
        assert self.instr.timeout == 10
        assert self.instr.get_visa_attribute(ResourceAttribute.timeout_value) == 10

        with pytest.raises(ValueError):
            self.instr.timeout = 10000000000

        del self.instr.timeout
        assert self.instr.timeout == float("+inf")
        assert (
            self.instr.get_visa_attribute(ResourceAttribute.timeout_value)
            == Timeouts.infinite
        )

    def test_resource_info(self):
        """Test accessing the resource info."""
        rinfo = self.instr.resource_info
        assert rinfo.interface_type == getattr(
            InterfaceType, self.rname.interface_type.lower()
        )

        assert rinfo.interface_board_number == int(self.rname.board)
        assert rinfo.resource_class == self.rname.resource_class
        assert rinfo.resource_name == str(self.rname)

    def test_interface_type(self):
        """Test accessing the resource interface_type."""
        assert self.instr.interface_type == getattr(
            InterfaceType, self.rname.interface_type.lower()
        )

    def test_attribute_handling(self):
        """Test directly manipulating attributes ie not using descriptors.

        This should extended in subclasses to test a broader range of
        attributes.

        """
        self.instr.set_visa_attribute(ResourceAttribute.timeout_value, 10)
        assert self.instr.get_visa_attribute(ResourceAttribute.timeout_value) == 10
        assert self.instr.timeout == 10

        self.instr.set_visa_attribute(
            ResourceAttribute.timeout_value, Timeouts.immediate
        )
        assert (
            self.instr.get_visa_attribute(ResourceAttribute.timeout_value)
            == self.MINIMAL_TIMEOUT
        )
        assert self.instr.timeout == 1

        self.instr.set_visa_attribute(
            ResourceAttribute.timeout_value, Timeouts.infinite
        )
        assert (
            self.instr.get_visa_attribute(ResourceAttribute.timeout_value)
            == Timeouts.infinite
        )
        assert self.instr.timeout == float("+inf")


class EventAwareResourceTestCaseMixin:
    """Mixing for resources supporting handling events."""

    def test_wait_on_event(self):
        """Test waiting on a VISA event.

        Should be implemented on subclasses, since the way to generate the
        event may be dependent on the resource type.

        """
        raise NotImplementedError()

    def test_managing_visa_handler(self):
        """Test using visa handlers.

        Should be implemented on subclasses, since the way to generate the
        event may be dependent on the resource type.

        """
        raise NotImplementedError()


class LockableResourceTestCaseMixin:
    """Mixing for resources supporting locking."""

    def test_shared_locking(self):
        """Test locking/unlocking a resource."""
        raise NotImplementedError()

    def test_exclusive_locking(self):
        """Test locking/unlocking a resource."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""ctypes wrapper for IVI-VISA library.

This file is part of PyVISA.

:copyright: 2014-2022 by PyVISA Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""
import os

from .highlevel import IVIVisaLibrary

WRAPPER_CLASS = IVIVisaLibrary

WRAP_HANDLER = True

env_value = os.environ.get("PYVISA_WRAP_HANDLER", None)
if env_value is not None:
    WRAP_HANDLER = bool(int(env_value))



# -*- coding: utf-8 -*-
"""VISA VPP-4.3 data types (VPP-4.3.2 spec, section 3) using ctypes constants.

This file is part of PyVISA.

All data types that are defined by VPP-4.3.2.

The module exports all data types including the pointer and array types.  This
means "ViUInt32" and such.

:copyright: 2014-2022 by PyVISA Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""
import ctypes as _ctypes
import sys

from .cthelper import FUNCTYPE

# Part One: Type Assignments for VISA and Instrument Drivers, see spec table
# 3.1.1.
#
# Remark: The pointer and probably also the array variants are of no
# significance in Python because there is no native call-by-reference.
# However, as long as I'm not fully sure about this, they won't hurt.


def _type_pair(ctypes_type):
    return ctypes_type, _ctypes.POINTER(ctypes_type)


def _type_triplet(ctypes_type):
    return _type_pair(ctypes_type) + (_ctypes.POINTER(ctypes_type),)


ViUInt64, ViPUInt64, ViAUInt64 = _type_triplet(_ctypes.c_uint64)
ViInt64, ViPInt64, ViAInt64 = _type_triplet(_ctypes.c_int64)
ViUInt32, ViPUInt32, ViAUInt32 = _type_triplet(_ctypes.c_uint32)
ViInt32, ViPInt32, ViAInt32 = _type_triplet(_ctypes.c_int32)
ViUInt16, ViPUInt16, ViAUInt16 = _type_triplet(_ctypes.c_ushort)
ViInt16, ViPInt16, ViAInt16 = _type_triplet(_ctypes.c_short)
ViUInt8, ViPUInt8, ViAUInt8 = _type_triplet(_ctypes.c_ubyte)
ViInt8, ViPInt8, ViAInt8 = _type_triplet(_ctypes.c_byte)
ViAddr, ViPAddr, ViAAddr = _type_triplet(_ctypes.c_void_p)
ViChar, ViPChar, ViAChar = _type_triplet(_ctypes.c_char)
ViByte, ViPByte, ViAByte = _type_triplet(_ctypes.c_ubyte)
ViBoolean, ViPBoolean, ViABoolean = _type_triplet(ViUInt16)
ViReal32, ViPReal32, ViAReal32 = _type_triplet(_ctypes.c_float)
ViReal64, ViPReal64, ViAReal64 = _type_triplet(_ctypes.c_double)


class ViString(object):
    @classmethod
    def from_param(cls, obj):
        if isinstance(obj, str):
            return bytes(obj, "ascii")
        return obj


class ViAString(object):
    @classmethod
    def from_param(cls, obj):
        return _ctypes.POINTER(obj)


ViPString = ViString

# This follows visa.h definition, but involves a lot of manual conversion.
# ViBuf, ViPBuf, ViABuf = ViPByte, ViPByte, _ctypes.POINTER(ViPByte)

ViBuf, ViPBuf, ViABuf = ViPString, ViPString, ViAString


def buffer_to_text(buf) -> str:
    return buf.value.decode("ascii")


ViRsrc = ViString
ViPRsrc = ViString
ViARsrc = ViAString

ViKeyId, ViPKeyId = ViString, ViPString

ViStatus, ViPStatus, ViAStatus = _type_triplet(ViInt32)
ViVersion, ViPVersion, ViAVersion = _type_triplet(ViUInt32)
_ViObject, ViPObject, ViAObject = _type_triplet(ViUInt32)
_ViSession, ViPSession, ViASession = _type_triplet(ViUInt32)


class ViObject(_ViObject):  # type: ignore
    @classmethod
    def from_param(cls, obj):
        if obj is None:
            raise ValueError("Session cannot be None. The resource might be closed.")
        return _ViObject.from_param(obj)


ViSession = ViObject


ViAttr = ViUInt32
ViConstString = _ctypes.POINTER(ViChar)


# Part Two: Type Assignments for VISA only, see spec table 3.1.2.  The
# difference to the above is of no significance in Python, so I use it here
# only for easier synchronisation with the spec.
is_64bit = sys.maxsize > 2**32

ViAccessMode, ViPAccessMode = _type_pair(ViUInt32)
ViBusAddress, ViPBusAddress = _type_pair(ViUInt64) if is_64bit else _type_pair(ViUInt32)
ViBusAddress64, ViPBusAddress64 = _type_pair(ViUInt64)

ViBusSize = ViUInt64 if is_64bit else ViUInt32
ViBusSize64 = ViUInt64

ViAttrState, ViPAttrState = _type_pair(ViUInt64) if is_64bit else _type_pair(ViUInt32)

# The following is weird, taken from news:zn2ek2w2.fsf@python.net
ViVAList = _ctypes.POINTER(_ctypes.c_char)

ViEventType, ViPEventType, ViAEventType = _type_triplet(ViUInt32)

ViPAttr = _ctypes.POINTER(ViAttr)
ViAAttr = ViPAttr

ViEventFilter = ViUInt32

ViFindList, ViPFindList = _type_pair(ViObject)
ViEvent, ViPEvent = _type_pair(ViObject)
ViJobId, ViPJobId = _type_pair(ViUInt32)

# Class of callback functions for event handling, first type is result type
ViHndlr = FUNCTYPE(ViStatus, ViSession, ViEventType, ViEvent, ViAddr)



# -*- coding: utf-8 -*-
"""VISA events with convenient access to the available attributes.

This file is part of PyVISA.

:copyright: 2020-2022 by PyVISA Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""
from typing import TYPE_CHECKING, Callable, Dict, Optional, Type

from typing_extensions import ClassVar

from . import attributes, constants, errors, logger
from .attributes import Attribute
from .typing import Any, VISAEventContext, VISAJobID

if TYPE_CHECKING:
    from . import highlevel  # pragma: no cover


class Event:
    """Event that lead to the call of an event handler.

    Do not instantiate directly use the visa_event context manager of the
    Resource object passed to the handler.

    Notes
    -----
    When using the queuing mechanism events are expected to be closed manually
    after handling them.
    When using callbacks, the events should only be closed if VISA will never
    get control again which since we always attempt to properly close all
    resources should never happend and hence events should never be closed manually.

    """

    #: Reference to the visa library
    visalib: "highlevel.VisaLibraryBase"

    #: Type of the event.
    event_type: constants.EventType

    #: Context use to query the event attributes.
    _context: Optional[VISAEventContext]

    #: Maps Event type to Python class encapsulating that event.
    _event_classes: ClassVar[Dict[constants.EventType, Type["Event"]]] = dict()

    @classmethod
    def register(
        cls, event_type: constants.EventType
    ) -> Callable[[Type["Event"]], Type["Event"]]:
        """Register a class with a given event type."""

        def _internal(python_class: Type["Event"]) -> Type["Event"]:
            if event_type in cls._event_classes:
                logger.warning(
                    "%s is already registered. "
                    "Overwriting with %s" % (event_type, python_class)
                )

            for attr in attributes.AttributesPerResource[event_type]:
                if not hasattr(python_class, attr.py_name):
                    raise TypeError(
                        "%s is expected to have an attribute %s"
                        % (python_class, attr.py_name)
                    )

            cls._event_classes[event_type] = python_class

            return python_class

        return _internal

    def __new__(
        cls,
        visalib: "highlevel.VisaLibraryBase",
        event_type: constants.EventType,
        context: VISAEventContext,
    ) -> "Event":
        event_cls = cls._event_classes.get(event_type, Event)
        return object.__new__(event_cls)

    def __init__(
        self,
        visalib: "highlevel.VisaLibraryBase",
        event_type: constants.EventType,
        context: Optional[VISAEventContext],
    ) -> None:
        self.visalib = visalib
        self.event_type = event_type
        self._context = context

    @property
    def context(self) -> VISAEventContext:
        """Access the VISA context used to retrieve attributes.

        This is equivalent to the session on a resource.

        """
        c = self._context
        if c is None:
            raise errors.InvalidSession()
        return c

    def get_visa_attribute(self, attribute_id: constants.EventAttribute) -> Any:
        """Get the specified VISA attribute."""
        return self.visalib.get_attribute(self.context, attribute_id)[0]

    def close(self):
        """Simply invalidate the context.

        The event is not closed since it is only ever required when using
        the queue mechanism and this is handled in WaitResponse.

        """
        self._context = None


# Those events do not have any payload beyond their type, for those use the base class
for event in (
    constants.EventType.clear,
    constants.EventType.service_request,
    constants.EventType.gpib_listen,
    constants.EventType.gpib_talk,
    constants.EventType.vxi_vme_sysfail,
    constants.EventType.vxi_vme_sysreset,
):
    Event.register(event)(Event)


@Event.register(constants.EventType.exception)
class ExceptionEvent(Event):
    """Event corresponding to an exception."""

    #: Status code of the operation that generated the exception
    status: Attribute[constants.StatusCode] = attributes.AttrVI_ATTR_STATUS()

    #: Name of the operation that led to the exception
    operation_name: Attribute[str] = attributes.AttrVI_ATTR_OPER_NAME()


@Event.register(constants.EventType.gpib_controller_in_charge)
class GPIBCICEvent(Event):
    """GPIB Controller in Charge event.

    The event is emitted if the status is gained or lost.

    """

    #: New state of the controller in charge status
    cic_state: Attribute[
        constants.LineState
    ] = attributes.AttrVI_ATTR_GPIB_RECV_CIC_STATE()


@Event.register(constants.EventType.io_completion)
class IOCompletionEvent(Event):
    """Event marking the completion of an IO operation."""

    #: Status code of the asynchronous I/O operation that has completed.
    status: Attribute[constants.StatusCode] = attributes.AttrVI_ATTR_STATUS()

    #: Buffer that was used in an asynchronous operation.
    buffer: Attribute[bytes] = attributes.AttrVI_ATTR_BUFFER()

    #: Actual number of elements that were asynchronously transferred.
    return_count: Attribute[int] = attributes.AttrVI_ATTR_RET_COUNT()

    #: Name of the operation generating the event.
    operation_name: Attribute[str] = attributes.AttrVI_ATTR_OPER_NAME()

    #: Job ID of the asynchronous operation that has completed.
    job_id: Attribute[VISAJobID] = attributes.AttrVI_ATTR_JOB_ID()

    @property
    def data(self):
        """Portion of the buffer that was actually filled during the call."""
        return bytes(self.buffer[: self.return_count])


@Event.register(constants.EventType.trig)
class TrigEvent(Event):
    """Trigger event."""

    #: Identifier of the triggering mechanism on which the specified trigger event
    #: was received.
    received_trigger_id: Attribute[
        constants.TriggerID
    ] = attributes.AttrVI_ATTR_TRIG_ID()


@Event.register(constants.EventType.usb_interrupt)
class USBInteruptEvent(Event):
    """USB interruption event."""

    #: Status of the read operation from the USB interrupt-IN pipe.
    status: Attribute[constants.StatusCode] = attributes.AttrVI_ATTR_STATUS()

    #: Size of the data that was received from the USB interrupt-IN pipe.
    size: Attribute[int] = attributes.AttrVI_ATTR_USB_RECV_INTR_SIZE()

    #: Actual data that was received from the USB interrupt-IN pipe.
    data = attributes.AttrVI_ATTR_USB_RECV_INTR_DATA()


@Event.register(constants.EventType.vxi_signal_interrupt)
class VXISignalInteruptEvent(Event):
    """VXI signal event."""

    #: 16-bit Status/ID value retrieved during the IACK cycle or
    #: from the Signal register.
    signal_register_status_id: Attribute[int] = attributes.AttrVI_ATTR_SIGP_STATUS_ID()


@Event.register(constants.EventType.vxi_vme_interrupt)
class VXIInterruptEvent(Event):
    """VXI interrupt event."""

    #: 32-bit status/ID retrieved during the IACK cycle.
    status_id: Attribute[int] = attributes.AttrVI_ATTR_INTR_STATUS_ID()

    #: VXI interrupt level on which the interrupt was received.
    level: Attribute[int] = attributes.AttrVI_ATTR_RECV_INTR_LEVEL()


@Event.register(constants.EventType.pxi_interrupt)
class PXIInteruptEvent(Event):
    """PXI interruption event."""

    #: Index of the interrupt sequence that detected the interrupt condition.
    sequence: Attribute[int] = attributes.AttrVI_ATTR_PXI_RECV_INTR_SEQ()

    #: First PXI/PCI register read in the successful interrupt detection sequence.
    data: Attribute[int] = attributes.AttrVI_ATTR_PXI_RECV_INTR_DATA()



# -*- coding: utf-8 -*-
"""Test pyvisa utility functions.

"""
import array
import contextlib
import logging
import os
import struct
import subprocess
import sys
import tempfile
import unittest
from configparser import ConfigParser
from functools import partial
from io import StringIO
from pathlib import Path
from types import ModuleType
from typing import Optional

import pytest

from pyvisa import highlevel, util
from pyvisa.ctwrapper import IVIVisaLibrary
from pyvisa.testsuite import BaseTestCase

np: Optional[ModuleType]
try:
    import numpy

    np = numpy
except ImportError:
    np = None


class TestConfigFile(BaseTestCase):
    """Test reading information from a user configuration file."""

    def setup_method(self):
        # Skip if a real config file exists
        if any(
            os.path.isfile(p)
            for p in [
                os.path.join(sys.prefix, "share", "pyvisa", ".pyvisarc"),
                os.path.join(os.path.expanduser("~"), ".pyvisarc"),
            ]
        ):
            raise unittest.SkipTest(
                ".pyvisarc file exists cannot properly test in this case"
            )
        self.temp_dir = tempfile.TemporaryDirectory()
        os.makedirs(os.path.join(self.temp_dir.name, "share", "pyvisa"))
        self.config_path = os.path.join(
            self.temp_dir.name, "share", "pyvisa", ".pyvisarc"
        )
        self._prefix = sys.prefix
        sys.prefix = self.temp_dir.name
        self._platform = sys.platform
        self._version_info = sys.version_info

    def teardown_method(self):
        self.temp_dir.cleanup()
        sys.prefix = self._prefix
        sys.platform = self._platform
        sys.version_info = self._version_info

    def test_reading_config_file(self):
        config = ConfigParser()
        config["Paths"] = {}
        config["Paths"]["visa library"] = "test"
        with open(self.config_path, "w") as f:
            config.write(f)
        assert util.read_user_library_path() == "test"

    def test_no_section(self, caplog):
        config = ConfigParser()
        with open(self.config_path, "w") as f:
            config.write(f)
        with caplog.at_level(level=logging.DEBUG):
            assert util.read_user_library_path() is None
        assert "NoOptionError or NoSectionError" in caplog.records[1].message

    def test_no_key(self, caplog):
        config = ConfigParser()
        config["Paths"] = {}
        with open(self.config_path, "w") as f:
            config.write(f)
        with caplog.at_level(level=logging.DEBUG):
            assert util.read_user_library_path() is None
        assert "NoOptionError or NoSectionError" in caplog.records[1].message

    def test_no_config_file(self, caplog):
        with caplog.at_level(level=logging.DEBUG):
            assert util.read_user_library_path() is None
        assert "No user defined" in caplog.records[0].message

    # --- Test reading dll_extra_paths.

    def test_reading_config_file_not_windows(self, caplog):
        sys.platform = "darwin"
        sys.version_info = (3, 8, 1)
        with caplog.at_level(level=logging.DEBUG):
            assert util.add_user_dll_extra_paths() is None
        assert "Not loading dll_extra_paths" in caplog.records[0].message

    def test_reading_config_file_old_python(self, caplog):
        sys.platform = "win32"
        sys.version_info = (3, 7, 1)
        with caplog.at_level(level=logging.DEBUG):
            assert util.add_user_dll_extra_paths() is None
        assert "Not loading dll_extra_paths" in caplog.records[0].message

    def test_reading_config_file_for_dll_extra_paths(self, monkeypatch):
        sys.platform = "win32"
        sys.version_info = (3, 8, 1)
        monkeypatch.setattr(
            os, "add_dll_directory", lambda *args, **kwargs: "", raising=False
        )
        config = ConfigParser()
        config["Paths"] = {}
        config["Paths"]["dll_extra_paths"] = r"C:\Program Files;C:\Program Files (x86)"
        with open(self.config_path, "w") as f:
            config.write(f)
        assert util.add_user_dll_extra_paths() == [
            r"C:\Program Files",
            r"C:\Program Files (x86)",
        ]

    def test_no_section_for_dll_extra_paths(self, monkeypatch, caplog):
        sys.platform = "win32"
        sys.version_info = (3, 8, 1)
        monkeypatch.setattr(
            os, "add_dll_directory", lambda *args, **kwargs: "", raising=False
        )
        config = ConfigParser()
        with open(self.config_path, "w") as f:
            config.write(f)
        with caplog.at_level(level=logging.DEBUG):
            assert util.add_user_dll_extra_paths() is None
        assert "NoOptionError or NoSectionError" in caplog.records[1].message

    def test_no_key_for_dll_extra_paths(self, monkeypatch, caplog):
        sys.platform = "win32"
        sys.version_info = (3, 8, 1)
        monkeypatch.setattr(
            os, "add_dll_directory", lambda *args, **kwargs: "", raising=False
        )
        config = ConfigParser()
        config["Paths"] = {}
        with open(self.config_path, "w") as f:
            config.write(f)
        with caplog.at_level(level=logging.DEBUG):
            assert util.add_user_dll_extra_paths() is None
        assert "NoOptionError or NoSectionError" in caplog.records[1].message

    def test_no_config_file_for_dll_extra_paths(self, monkeypatch, caplog):
        sys.platform = "win32"
        sys.version_info = (3, 8, 1)
        monkeypatch.setattr(
            os, "add_dll_directory", lambda *args, **kwargs: "", raising=False
        )
        with caplog.at_level(level=logging.DEBUG):
            assert util.add_user_dll_extra_paths() is None
        assert "No user defined" in caplog.records[0].message


class TestParser(BaseTestCase):
    def test_parse_binary(self):
        s = (
            b"#0@\xe2\x8b<@\xe2\x8b<@\xe2\x8b<@\xe2\x8b<@\xde\x8b<@\xde\x8b<@"
            b"\xde\x8b<@\xde\x8b<@\xe0\x8b<@\xe0\x8b<@\xdc\x8b<@\xde\x8b<@"
            b"\xe2\x8b<@\xe0\x8b<"
        )
        e = [
            0.01707566,
            0.01707566,
            0.01707566,
            0.01707566,
            0.01707375,
            0.01707375,
            0.01707375,
            0.01707375,
            0.01707470,
            0.01707470,
            0.01707280,
            0.01707375,
            0.01707566,
            0.01707470,
        ]

        # Test handling indefinite length block
        p = util.from_ieee_block(s, datatype="f", is_big_endian=False)
        for a, b in zip(p, e):
            assert a == pytest.approx(b)

        # Test handling definite length block
        p = util.from_ieee_block(b"#214" + s[2:], datatype="f", is_big_endian=False)
        for a, b in zip(p, e):
            assert a == pytest.approx(b)

        # Test handling zero length block
        p = util.from_ieee_block(b"#10" + s[2:], datatype="f", is_big_endian=False)
        assert not p

        p = util.from_hp_block(
            b"#A\x0e\x00" + s[2:],
            datatype="f",
            is_big_endian=False,
            container=partial(array.array, "f"),
        )
        for a, b in zip(p, e):
            assert a == pytest.approx(b)

    def test_integer_ascii_block(self):
        values = list(range(99))
        for fmt in "d":
            msg = "block=%s, fmt=%s"
            msg = msg % ("ascii", fmt)
            # Test handling the case of a trailing comma
            tb = lambda values: util.to_ascii_block(values, fmt, ",") + ","
            fb = lambda block, cont: util.from_ascii_block(block, fmt, ",", cont)
            self.round_trip_block_conversion(values, tb, fb, msg)

    def test_non_integer_ascii_block(self):
        values = [val + 0.5 for val in range(99)]
        values = list(range(99))
        for fmt in "fFeEgG":
            msg = "block=%s, fmt=%s"
            msg = msg % ("ascii", fmt)
            tb = lambda values: util.to_ascii_block(values, fmt, ",")
            fb = lambda block, cont: util.from_ascii_block(block, fmt, ",", cont)
            self.round_trip_block_conversion(values, tb, fb, msg)

    def test_invalid_string_converter(self):
        with pytest.raises(ValueError) as ex:
            util.to_ascii_block([1, 2], "m")
        assert "unsupported format character" in ex.exconly()
        with pytest.raises(ValueError) as ex:
            util.from_ascii_block("1,2,3", "m")
        assert "Invalid code for converter" in ex.exconly()

    def test_function_separator(self):
        values = list(range(99))
        fmt = "d"
        msg = "block=ascii, fmt=%s" % fmt
        tb = lambda values: util.to_ascii_block(values, fmt, ":".join)
        fb = lambda block, cont: util.from_ascii_block(
            block, fmt, lambda s: s.split(":"), cont
        )
        self.round_trip_block_conversion(values, tb, fb, msg)

    def test_function_converter(self):
        values = list(range(99))
        msg = "block=ascii"
        tb = lambda values: util.to_ascii_block(values, str, ":".join)
        fb = lambda block, cont: util.from_ascii_block(
            block, int, lambda s: s.split(":"), cont
        )
        self.round_trip_block_conversion(values, tb, fb, msg)

    def test_integer_binary_block(self):
        values = list(range(99))
        for block, tb, fb in zip(
            ("ieee", "hp"),
            (util.to_ieee_block, util.to_hp_block),
            (util.from_ieee_block, util.from_hp_block),
        ):
            for fmt in "bBhHiIfd":
                for endi in (True, False):
                    msg = "block=%s, fmt=%s, endianness=%s"
                    msg = msg % (block, fmt, endi)
                    tblock = lambda values: tb(values, fmt, endi)
                    fblock = lambda block, cont: fb(block, fmt, endi, cont)
                    self.round_trip_block_conversion(values, tblock, fblock, msg)

    def test_noninteger_binary_block(self):
        values = [val + 0.5 for val in range(99)]
        for block, tb, fb in zip(
            ("ieee", "hp"),
            (util.to_ieee_block, util.to_hp_block),
            (util.from_ieee_block, util.from_hp_block),
        ):
            for fmt in "fd":
                for endi in (True, False):
                    msg = "block=%s, fmt=%s, endianness=%s"
                    msg = msg % (block, fmt, endi)
                    tblock = lambda values: bytearray(tb(values, fmt, endi))
                    fblock = lambda block, cont: fb(block, fmt, endi, cont)
                    self.round_trip_block_conversion(values, tblock, fblock, msg)

    def test_bytes_binary_block(self):
        values = b"dbslbw cj saj \x00\x76"
        for block, tb, fb in zip(
            ("ieee", "hp"),
            (util.to_ieee_block, util.to_hp_block),
            (util.from_ieee_block, util.from_hp_block),
        ):
            for fmt in "sbB":
                block = tb(values, datatype=fmt)
                print(fmt, block)
                rt = fb(block, datatype=fmt, container=bytes)
                assert values == rt

    def test_malformed_binary_block_header(self):
        values = list(range(10))
        for header, tb, fb in zip(
            ("ieee", "hp"),
            (util.to_ieee_block, util.to_hp_block),
            (util.from_ieee_block, util.from_hp_block),
        ):
            block = tb(values, "h", False)
            bad_block = block[1:]
            with pytest.raises(ValueError) as e:
                fb(bad_block, "h", False, list)

            assert "(#" in e.exconly()

    def test_weird_binary_block_header(self):
        values = list(range(100))
        for header, tb, fb in zip(
            ("ieee", "hp"),
            (util.to_ieee_block, util.to_hp_block),
            (util.from_ieee_block, util.from_hp_block),
        ):
            block = tb(values, "h", False)
            bad_block = block[1:]
            if header == "hp":
                index = bad_block.find(b"#")
                bad_block = bad_block[:index] + b"#A" + bad_block[index + 2 :]
            with pytest.warns(UserWarning):
                fb(bad_block, "h", False, list)

    def test_weird_binary_block_header_raise(self):
        values = list(range(100))
        for header, tb, fb in zip(
            ("ieee", "hp"),
            (util.to_ieee_block, util.to_hp_block),
            (util.from_ieee_block, util.from_hp_block),
        ):
            block = tb(values, "h", False)
            bad_block = block[1:]
            if header == "hp":
                index = bad_block.find(b"#")
                bad_block = bad_block[:index] + b"#A" + bad_block[index + 2 :]
            parse = (
                util.parse_ieee_block_header
                if header == "ieee"
                else partial(util.parse_hp_block_header, is_big_endian=False)
            )

            with pytest.raises(RuntimeError):
                parse(bad_block, raise_on_late_block=True)

            parse(bad_block, length_before_block=1000)

    def test_binary_block_shorter_than_advertized(self):
        values = list(range(99))
        for header, tb, fb in zip(
            ("ieee", "hp"),
            (util.to_ieee_block, util.to_hp_block),
            (util.from_ieee_block, util.from_hp_block),
        ):
            block = tb(values, "h", False)
            if header == "ieee":
                header_byte_number = int(block[1])
                block = (
                    block[:2]
                    + b"9" * header_byte_number
                    + block[2 + header_byte_number :]
                )
            else:
                block = block[:2] + b"\xff\xff\xff\xff" + block[2 + 4 :]
            with pytest.raises(ValueError) as e:
                fb(block, "h", False, list)

            assert "Binary data is incomplete" in e.exconly()

    def test_guessing_block_length(self):
        values = list(range(99))
        for header, tb, fb in zip(
            ("ieee", "hp"),
            (util.to_ieee_block, util.to_hp_block),
            (util.from_ieee_block, util.from_hp_block),
        ):
            block = tb(values, "h", False) + b"\n"
            if header == "ieee":
                header_length = int(block[1:2].decode())
                block = block[:2] + b"0" * header_length + block[2 + header_length :]
            else:
                block = block[:2] + b"\x00\x00\x00\x00" + block[2 + 4 :]
            assert not fb(block, "h", False, list)

    def test_handling_malformed_binary(self):
        containers = (list, tuple) + ((np.array, np.ndarray) if np else ())

        # Use this to generate malformed data which should in theory be
        # impossible
        class DumbBytes(bytes):
            def __len__(self):
                return 10

        for container in containers:
            with pytest.raises(ValueError) as e:
                util.from_binary_block(DumbBytes(b"\x00\x00\x00"), container=container)
            assert (
                "malformed" if container in (list, tuple) else "buffer" in e.exconly()
            )

    def round_trip_block_conversion(self, values, to_block, from_block, msg):
        """Test that block conversion round trip as expected."""
        containers = (list, tuple) + ((np.array,) if np else ())
        for cont in containers:
            conv = cont(values)
            msg += ", container=%s"
            msg = msg % cont.__name__
            try:
                block = to_block(conv)
                parsed = from_block(block, cont)
            except Exception as e:
                raise Exception(msg + "\n" + repr(e))

            if np and cont in (np.array,):
                np.testing.assert_array_equal(conv, parsed, msg)
            else:
                assert conv == parsed, msg


class TestSystemDetailsAnalysis(BaseTestCase):
    """Test getting the system details."""

    def setup_method(self):
        self._unicode_size = sys.maxunicode

    def teardown_method(self):
        sys.maxunicode = self._unicode_size

    def test_getting_system_details(self):
        sys.maxunicode = 65535
        path = os.path.join(os.path.dirname(__file__), "fake-extensions")
        sys.path.append(path)
        try:
            details = util.get_system_details(True)
        finally:
            sys.path.remove(path)
        assert details["backends"]
        assert details["unicode"] == "UCS2"

        sys.maxunicode = 1114111
        details = util.get_system_details(False)
        assert not details["backends"]
        assert details["unicode"] == "UCS4"

    def test_get_debug_info(self):
        details = util.system_details_to_str(util.get_system_details())
        assert util.get_debug_info(False) == details
        temp_stdout = StringIO()
        with contextlib.redirect_stdout(temp_stdout):
            util.get_debug_info()
        output = temp_stdout.getvalue()
        assert output.strip() == details.strip()

    def test_system_details_for_plugins(self):
        """Test reporting on plugins."""

        def dummy_list_backends():
            return ["test1", "test2", "test3", "test4"]

        def dummy_get_wrapper_class(backend):
            if backend == "test1":
                return IVIVisaLibrary

            elif backend == "test2":

                class BrokenBackend:
                    @classmethod
                    def get_debug_info(cls):
                        raise Exception()

                return BrokenBackend

            elif backend == "test4":

                class WeirdBackend:
                    @classmethod
                    def get_debug_info(cls):
                        return {"": {"": [object()]}}

                return WeirdBackend

            else:
                raise Exception()

        old_lb = highlevel.list_backends
        old_gwc = highlevel.get_wrapper_class
        highlevel.list_backends = dummy_list_backends
        highlevel.get_wrapper_class = dummy_get_wrapper_class

        try:
            details = util.get_system_details()
        finally:
            highlevel.list_backends = old_lb
            highlevel.get_wrapper_class = old_gwc

        assert "Could not instantiate" in details["backends"]["test3"][0]
        assert "Could not obtain" in details["backends"]["test2"][0]
        assert "Version" in details["backends"]["test1"]
        assert "" in details["backends"]["test4"]

        # Test converting the details to string
        util.system_details_to_str(details)


def generate_fakelibs(dirname: Path):
    libs = {
        "fakelib_bad_magic.dll": struct.pack(
            "=6sH52sl", b"MAPE\x00\x00", 0x014C, 52 * b"\0", 2
        ),
        "fakelib_good_x86_32.dll": struct.pack(
            "=6sH52sl", b"MZPE\x00\x00", 0x014C, 52 * b"\0", 2
        ),
        "fakelib_good_x86_64.dll": struct.pack(
            "=6sH52sl", b"MZPE\x00\x00", 0x8664, 52 * b"\0", 2
        ),
        "fakelib_good_arm_64.dll": struct.pack(
            "=6sH52sl", b"MZPE\x00\x00", 0xAA64, 52 * b"\0", 2
        ),
        "fakelib_good_unknown.dll": struct.pack(
            "=6sH52sl", b"MZPE\x00\x00", 0xFFFF, 52 * b"\0", 2
        ),
        "fakelib_not_pe.dll": struct.pack(
            "=6sH52sl", b"MZDE\x00\x00", 0x014C, 52 * b"\0", 2
        ),
    }
    for name, blob in libs.items():
        with open(dirname / name, "wb") as f:
            f.write(blob)
            print("Written %s" % name)


class TestLibraryAnalysis(BaseTestCase):
    """Test (through monkey patching) the analysis of binary libraries."""

    def test_get_shared_library_arch(self, tmp_path: Path):
        """Test analysing a library on Windows."""
        generate_fakelibs(tmp_path)

        for f, a in zip(
            ["x86_32", "x86_64", "arm_64"],
            [
                util.PEMachineType.I386,
                util.PEMachineType.AMD64,
                util.PEMachineType.AARCH64,
            ],
        ):
            arch = util.get_shared_library_arch(tmp_path / f"fakelib_good_{f}.dll")
            assert arch == a

        arch = util.get_shared_library_arch(tmp_path / "fakelib_good_unknown.dll")
        assert arch == util.PEMachineType.UNKNOWN

        with pytest.raises(Exception) as e:
            util.get_shared_library_arch(tmp_path / "fakelib_bad_magic.dll")
        assert "Not an executable" in e.exconly()

        with pytest.raises(Exception) as e:
            util.get_shared_library_arch(tmp_path / "fakelib_not_pe.dll")
        assert "Not a PE executable" in e.exconly()

    def test_get_arch_windows(self, tmp_path: Path):
        """Test identifying the computer architecture on windows."""
        generate_fakelibs(tmp_path)

        platform = sys.platform
        sys.platform = "win32"
        try:
            for f, a in zip(
                ["x86_32", "x86_64", "arm_64", "_unknown"],
                [
                    [util.ArchitectureType.I386],
                    [util.ArchitectureType.X86_64],
                    [util.ArchitectureType.AARCH64],
                    [],
                ],
            ):
                print(f, a)
                path = tmp_path / f"fakelib_good_{f}.dll"
                lib = util.LibraryPath(str(path))
                assert lib.arch == a
        finally:
            sys.platform = platform

    @pytest.mark.skipif(sys.version_info < (3, 7), reason="Fails weirdly on Python 3.6")
    def test_get_arch_unix(self):
        """Test identifying the computer architecture on linux and Mac."""
        platform = sys.platform
        run = subprocess.run
        try:

            def alt_run(*args, **kwargs):
                if platform.startswith("win"):
                    kwargs["shell"] = True
                return run(["echo", args[0][1]], *args[1:], **kwargs)

            subprocess.run = alt_run

            for p, f, a in [
                ("linux", "80386", [(util.ArchitectureType.I386)]),
                ("linux", "x86-64", [(util.ArchitectureType.X86_64)]),
                ("linux", "aarch64", [(util.ArchitectureType.AARCH64)]),
                ("darwin", "executable i386", [(util.ArchitectureType.I386)]),
                ("darwin", "executable x86_64", [(util.ArchitectureType.X86_64)]),
                ("darwin", "executable arm64", [(util.ArchitectureType.AARCH64)]),
            ]:
                sys.platform = p
                lib = util.LibraryPath(f)
                assert lib.arch == a
        finally:
            sys.platform = platform
            subprocess.run = run

    def test_get_arch_unix_unreported(self):
        """Test identifying the computer architecture on an unknown platform."""
        platform = sys.platform
        run = subprocess.run
        try:
            sys.platform = "darwin"
            lib = util.LibraryPath("")
            assert lib.arch == []
        finally:
            sys.platform = platform
            subprocess.run = run

    def test_get_arch_unknown(self):
        """Test identifying the computer architecture on an unknown platform."""
        platform = sys.platform
        run = subprocess.run
        try:
            sys.platform = "test"
            lib = util.LibraryPath("")
            assert lib.arch == []
        finally:
            sys.platform = platform
            subprocess.run = run
